import os
import sharedmem
import ctypes

import numpy as np
import numpy.testing as npt
import nose.tools as nt
from scipy.signal import fftconvolve

import popeye.utilities as utils
import popeye.og as og
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus
from popeye.spinach import generate_og_receptive_field

def test_og_fit():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 6
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5    
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-10,10,5)
    y_grid = utils.grid_slice(-10,10,5)
    s_grid = utils.grid_slice (0.25,5.25,5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # fit the response
    fit = og.GaussianFit(model, data, grids, bounds)
    
    # coarse fit
    ballpark = [-5.       ,  5.       ,  2.75]
    
    npt.assert_almost_equal((fit.x0, fit.y0, fit.s0), ballpark)
    # the baseline/beta should be 0/1 when regressed data vs. estimate
    (m,b) = np.polyfit(fit.scaled_ballpark_prediction, data, 1)
    npt.assert_almost_equal(m, 1.0)
    npt.assert_almost_equal(b, 0.0)
    
    # assert equivalence
    npt.assert_almost_equal(fit.x, x)
    npt.assert_almost_equal(fit.y, y)
    npt.assert_almost_equal(fit.sigma, sigma)
    npt.assert_almost_equal(fit.beta, beta)
    
    # test receptive field
    rf = generate_og_receptive_field(x, y, sigma, fit.model.stimulus.deg_x, fit.model.stimulus.deg_y)
    rf /= (2 * np.pi * sigma**2) * 1/np.diff(model.stimulus.deg_x[0,0:2])**2
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field.sum())) 
    
    # test model == fit RF
    npt.assert_almost_equal(np.round(fit.model.generate_receptive_field(x,y,sigma).sum()), np.round(fit.receptive_field.sum()))

def test_negative_og_fit():

    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf, utils.percent_change)
    model.hrf_delay = 0
    model.mask_size = 6
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = -0.25
    baseline = 0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-10,10,5)
    y_grid = utils.grid_slice(-10,10,5)
    s_grid = utils.grid_slice (0.25,5.25,5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (None,None)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # fit the response
    fit = og.GaussianFit(model, data, grids, bounds)
    
    # coarse fit
    ballpark = [-5.0, 5.0, 2.75]
    
    npt.assert_almost_equal((fit.x0, fit.y0, fit.s0), ballpark)
    # the baseline/beta should be 0/1 when regressed data vs. estimate
    (m,b) = np.polyfit(fit.scaled_ballpark_prediction, data, 1)
    npt.assert_almost_equal(m, 1.0)
    npt.assert_almost_equal(b, 0.0)
    
    # assert equivalence
    npt.assert_almost_equal(fit.x, x, 2)
    npt.assert_almost_equal(fit.y, y, 2)
    npt.assert_almost_equal(fit.sigma, sigma, 2)
    npt.assert_almost_equal(fit.beta, beta, 2)
    
    nt.assert_false(fit.model.bounded_amplitude)
    nt.assert_true(fit.slope<0)
    nt.assert_true(fit.beta0<0)
    nt.assert_true(fit.beta<0)
    
    # test receptive field
    rf = generate_og_receptive_field(x, y, sigma, fit.model.stimulus.deg_x, fit.model.stimulus.deg_y)
    rf /= (2 * np.pi * sigma**2) * 1/np.diff(model.stimulus.deg_x[0,0:2])**2
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field.sum())) 
    
    # test model == fit RF
    npt.assert_almost_equal(np.round(fit.model.generate_receptive_field(x,y,sigma).sum()), np.round(fit.receptive_field.sum()))

def test_bounded_amplitude_failure():

    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16

    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf, utils.percent_change)
    model.hrf_delay = 0
    model.mask_size = 6
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = -0.25
    baseline = 0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-10,10,5)
    y_grid = utils.grid_slice(-10,10,5)
    s_grid = utils.grid_slice (0.25,5.25,5)

    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)

    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)

    # fit the response
    fit = og.GaussianFit(model, data, grids, bounds)
    
    nt.assert_true(fit.model.bounded_amplitude)
    nt.assert_true(fit.slope>0)
    nt.assert_true(fit.beta0>0)
    nt.assert_true(fit.beta>0)
    nt.assert_true(fit.beta != beta)
    
# def test_og_nuisance_fit():
#     
#     # stimulus features
#     viewing_distance = 38
#     screen_width = 25
#     thetas = np.arange(0,360,90)
#     thetas = np.insert(thetas,0,-1)
#     thetas = np.append(thetas,-1)
#     num_blank_steps = 30
#     num_bar_steps = 30
#     ecc = 12
#     tr_length = 1.0
#     frames_per_tr = 1.0
#     scale_factor = 1.0
#     pixels_across = 200
#     pixels_down = 200
#     dtype = ctypes.c_int16
#     
#     # create the sweeping bar stimulus in memory
#     bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
#                                 screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
#                                 
#     # create an instance of the Stimulus class
#     stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
#     
#     # initialize the gaussian model
#     model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
#     model.hrf_delay = 0
#     
#     # generate a random pRF estimate
#     x = -5.24
#     y = 2.58
#     sigma = 0.98
#     beta = 2.5
#     baseline = 0.0
#     
#     # create the "data"
#     data = model.generate_prediction(x, y, sigma, beta, baseline)
#     
#     # create nuisance signal
#     step = np.zeros(len(data))
#     step[30:-30] = 1
#     
#     # add to data
#     data += step
#     
#     # create design matrix
#     nuisance = sm.add_constant(step)
#     
#     # recreate model with nuisance
#     model = og.GaussianModel(stimulus, utils.double_gamma_hrf, nuisance)
#     model.hrf_delay = 0
#     
#     # set search grid
#     x_grid = (-7,7)
#     y_grid = (-7,7)
#     s_grid = (0.25,3.25)
#     
#     # set search bounds
#     x_bound = (-10.0,10.0)
#     y_bound = (-10.0,10.0)
#     s_bound = (0.001,10.0)
#     b_bound = (1e-8,None)
#     m_bound = (None, None)
#     
#     # loop over each voxel and set up a GaussianFit object
#     grids = (x_grid, y_grid, s_grid,)
#     bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
#     
#     # fit the response
#     fit = og.GaussianFit(model, data, grids, bounds, Ns=3)
#     
#     # assert equivalence
#     nt.assert_almost_equal(fit.x, x, 1)
#     nt.assert_almost_equal(fit.y, y, 1)
#     nt.assert_almost_equal(fit.sigma, sigma, 1)
#     nt.assert_almost_equal(fit.beta, beta, 1)
    



from __future__ import division
from random import shuffle
from itertools import repeat
import time
import ctypes
import numpy as np

from scipy import ndimage
from scipy.optimize import fmin_powell, fmin
import nibabel

from popeye import og
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus
from popeye.spinach import generate_og_receptive_field, generate_og_receptive_fields


def recast_simulation_results(output, grid_parent):
    
    # load the gridParent
    dims = list(grid_parent.shape)
    dims = dims[0:3]
    dims.append(3)
    
    # initialize the statmaps
    estimates = np.zeros(dims)
    
    # extract the prf model estimates from the results queue output
    for o in output:
        voxel_index = o[0]
        estimates[voxel_index] = (o[1],float(o[2][0]),o[3])

        
    # get header information from the gridParent and update for the prf volume
    aff = grid_parent.get_affine()
    hdr = grid_parent.get_header()
    hdr.set_data_shape(dims)
    
    # recast as nifti
    nifti_estimates = nibabel.Nifti1Image(estimates,aff,header=hdr)
    
    return nifti_estimates

def error_function(neural_sigma, voxel_sigma, voxel_rf, deg_x, deg_y, xs, ys):
    
    if neural_sigma <= 0:
        return np.inf
    if neural_sigma > voxel_sigma:
        return np.inf
    
    # create all the neural rfs
    neural_rfs = generate_og_receptive_fields(deg_x, deg_y, xs, ys, neural_sigma)
    
    # normalize each rf by integral
    neural_rfs /= 2 * np.pi * neural_sigma ** 2
    
    # sum and divide by the number of neurons
    neural_rf = np.sum(neural_rfs,axis=-1)/neural_rfs.shape[-1]
    
    # RSS between neural and voxel
    error = np.sum((neural_rf-voxel_rf)**2)
    
    return error

def simulate_neural_sigma(estimate, scatter, deg_x, deg_y, voxel_index, num_neurons=1000, verbose=True):
    
    # timestamp
    start = time.clock()
    
    # unpack
    x = estimate[0]
    y = estimate[1]
    sigma = estimate[2]
    
    # create the Gaussian
    voxel_rf = generate_og_receptive_field(deg_x, deg_y, x, y, sigma)
    
    # normalize by integral
    voxel_rf /= 2*np.pi*sigma**2
    
    # generate random angles and scatters
    angles = np.random.uniform(0,2*np.pi,num_neurons)
    lengths = np.random.uniform(0,scatter,num_neurons)
    
    # convert to cartesian coords
    xs = x + np.sin(angles)*lengths
    ys = y + np.cos(angles)*lengths
    
    sigma_phat = fmin_powell(error_function, sigma, args=(sigma, voxel_rf, deg_x, deg_y, xs, ys),full_output=True,disp=False)
    
    # timestamp
    finish = time.clock()
    
    # progress
    if verbose:
        txt = ("VOXEL=(%.03d,%.03d,%.03d)   TIME=%.03d   OLD=%.02f  NEW=%.02f   SCATTER=%.02f" 
            %(voxel_index[0],
              voxel_index[1],
              voxel_index[2],
              finish-start,
              sigma,
              sigma_phat[0],
              scatter))
        
        print(txt)
    
    return (voxel_index,sigma,sigma_phat,scatter)

def parallel_simulate_neural_sigma(args):
    
    estimate = args[0]
    scatter = args[1]
    model = args[2]
    voxel_index = args[3]
    num_neurons = args[4]
    
    neural_sigma = simulate_neural_sigma(estimate, scatter, model.stimulus.deg_x, model.stimulus.deg_y, voxel_index, num_neurons)
    
    return neural_sigma

def generate_scatter_volume(roi, prf, threshold, iterations):
    
    scatter = np.zeros_like(roi,dtype='double')
    
    # get indices
    xi,yi,zi = np.nonzero((roi>0) & (prf[...,-3]>threshold))
    
    for voxel in range(len(xi)):
        
        # get index
        xind = xi[voxel]
        yind = yi[voxel]
        zind = zi[voxel]
        
        # set target to 1
        mask = np.zeros_like(roi)
        mask[xind,yind,zind] = 1
        
        # get neighborhood
        hood = ndimage.binary_dilation(mask,iterations=iterations)
        hood[xind,yind,zind] = 0
        
        fit = prf[mask==1][0]
        fits = prf[hood==1]
        scatter[xind,yind,zind] = np.mean(np.sqrt((fits[:,0]-fit[np.newaxis,0])**2 + (fits[:,1]-fit[np.newaxis,1])**2))/2
    
    return scatter


"""

First pass at a stimulus model for abstracting the qualities and functionality of a stimulus
into an abstract class.  For now, we'll assume the stimulus model only pertains to visual 
stimuli on a visual display over time (i.e., 3D).  Hopefully this can be extended to other stimuli
with an arbitrary number of dimensions (e.g., auditory stimuli).

"""

from __future__ import division
import ctypes
import gc
import sys  

import numpy as np
from scipy.ndimage.interpolation import zoom
from scipy.io import loadmat
from scipy.signal import square

from popeye.base import StimulusModel
import popeye.utilities as utils

def pixels_per_degree(pixels_across, screen_width, viewing_distance):

    return np.pi*pixels_across/np.arctan(screen_width/viewing_distance/2.0)/360.0
    
def generate_coordinate_matrices(pixels_across, pixels_down, ppd, scale_factor=1):
    
    """Creates coordinate matrices for representing the visual field in terms
       of degrees of visual angle.
       
    This function takes the screen dimensions, the pixels per degree, and a
    scaling factor in order to generate a pair of ndarrays representing the
    horizontal and vertical extents of the visual display in degrees of visual
    angle.
    
    Parameters
    ----------
    pixels_across : int
        The number of pixels along the horizontal extent of the visual display.
    pixels_down : int
        The number of pixels along the vertical extent of the visual display.
    ppd: float
        The number of pixels that spans 1 degree of visual angle.  This number
        is computed using the display width and the viewing distance.  See the
        config.init_config for details. 
    scale_factor : float
        The scale factor by which the stimulus is resampled.  The scale factor
        must be a float, and must be greater than 0.
        
    Returns
    -------
    deg_x : ndarray
        An array representing the horizontal extent of the visual display in
        terms of degrees of visual angle.
    deg_y : ndarray
        An array representing the vertical extent of the visual display in
        terms of degrees of visual angle.
    """
    
    [X,Y] = np.meshgrid(np.arange(np.round(pixels_across*scale_factor)),
                        np.arange(np.round(pixels_down*scale_factor)))
                        
                        
    deg_x = (X-np.round(pixels_across*scale_factor)/2)/(ppd*scale_factor)
    deg_y = (Y-np.round(pixels_down*scale_factor)/2)/(ppd*scale_factor)
    
    deg_x += 0.5/(ppd*scale_factor)
    deg_y += 0.5/(ppd*scale_factor)
    
    return deg_x, np.flipud(deg_y)

def resample_stimulus(stim_arr, scale_factor=0.05, mode='nearest',
                      order=0, dtype='uint8'):
    
    """Resamples the visual stimulus
    
    The function takes an ndarray `stim_arr` and resamples it by the user
    specified `scale_factor`.  The stimulus array is assumed to be a three
    dimensional ndarray representing the stimulus, in screen pixel coordinates,
    over time.  The first two dimensions of `stim_arr` together represent the
    exent of the visual display (pixels) and the last dimensions represents
    time (TRs).

    The underlying function used here is `scipy.ndimage.zoom`. Some arguments
    are passed through to that function.
    
    Parameters
    ----------
    stim_arr : ndarray
        Array_like means all those objects -- lists, nested lists, etc. --
        that can be converted to an array.
    
    scale_factor : float
        The scale factor by which the stimulus is resampled.  The scale factor
        must be a float, and must be greater than 0.
    
    mode : str, optional
        Points outside the boundaries of the input are filled according
        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
        Default is 'nearest'.

    order : int, optional
        Interpolation order, must be in range 0-5.

    dtype : numpy dtype, optional
        Datatype for the returned array.
        
    Returns
    -------
    resampled_arr : ndarray
        An array that is resampled according to the user-specified scale factor.
    """
    
    dims = np.shape(stim_arr)
    resampled_arr = np.zeros((int(round(dims[0] * scale_factor)),
                              int(round(dims[1] * scale_factor)),
                              dims[2]), dtype=dtype)
    
    # loop
    for tr in np.arange(dims[-1]):
        
        # resize it
        f = zoom(stim_arr[:,:,tr], scale_factor, mode=mode, order=order)
        
        # insert it
        resampled_arr[:,:,tr] = f
    
    return resampled_arr

def gaussian_2D(X, Y, x0, y0, sigma_x, sigma_y, degrees, amplitude=1):
    
    """
    A utility function for creating a two-dimensional Gaussian.
    
    This function served as the model for the Cython implementation of a
    generator of two-dimensional Gaussians.
    
    X : ndarray
        The coordinate array for the horizontal dimension of the display.
        
    Y : ndarray
        The coordinate array for the the vertical dimension of the display.
    
    x0 : float
        The location of the center of the Gaussian on the horizontal axis.
        
    y0 : float
        The location of the center of the Gaussian on the verticala axis.
    
    sigma_x : float
        The dispersion of the Gaussian over the horizontal axis.
    
    sigma_y : float
        The dispersion of the Gaussian over the vertical axis.
    
    degrees : float
        The orientation of the two-dimesional Gaussian (degrees).
    
    amplitude : float
        The amplitude of the two-dimensional Gaussian.
    
    Returns
    -------
    gauss : ndarray
        The two-dimensional Gaussian.
    
    """
    theta = degrees*np.pi/180
        
    a = np.cos(theta)**2/2/sigma_x**2 + np.sin(theta)**2/2/sigma_y**2
    b = -np.sin(2*theta)/4/sigma_x**2 + np.sin(2*theta)/4/sigma_y**2
    c = np.sin(theta)**2/2/sigma_x**2 + np.cos(theta)**2/2/sigma_y**2
    
    Z = amplitude*np.exp( - (a*(X-x0)**2 + 2*b*(X-x0)*(Y-y0) + c*(Y-y0)**2))
    
    return Z

def simulate_sinflicker_bar(pixels_across, pixels_down, 
                            viewing_distance, screen_width, 
                            thetas, sweep_steps, bar_width,
                            ecc, tr_length, flicker_hz, projector_hz):
    
    # get number of frames per volume
    frames_per_vol = tr_length*projector_hz
    total_trs = len(thetas)*sweep_steps
    total_frames = frames_per_vol * total_trs
    total_secs = total_trs*tr_length
    
    # flicker
    t = np.linspace(0,total_secs,total_secs*projector_hz)
    full_run = np.sin(2 * np.pi * flicker_hz * t)
    full_run = np.uint8((full_run + 1) * 128)
    
    # visuotopic stuff
    ppd = np.pi*pixels_across/np.arctan(screen_width/viewing_distance/2.0)/360.0 # degrees of visual angle
    deg_x, deg_y = generate_coordinate_matrices(pixels_across, pixels_down, ppd, 1.0)
    
    # initialize the bar array
    bar_stimulus = np.uint8(np.ones((pixels_down, pixels_across, total_frames)))
    
    # counter
    tr_num = 0
    
    # main loop
    for theta in thetas:
        
        if theta != -1:
            
            # convert to radians
            theta_rad = theta * np.pi / 180
            
            # get the starting point and trajectory
            start_pos = np.array([-np.cos(theta_rad)*ecc, -np.sin(theta_rad)*ecc])
            end_pos = np.array([np.cos(theta_rad)*ecc, np.sin(theta_rad)*ecc])
            run_and_rise = end_pos - start_pos;
            
            if np.mod(theta,90) == 0:
                sigma_x = bar_width/2
                sigma_y = 500
            else:
                sigma_x = 500 # pragma: no cover
                sigma_y = bar_width/2 # pragma: no cover
            
            # step through each position along the trajectory
            for step in np.arange(0,sweep_steps):
                
                # get the position of the bar at each step
                xy0 = run_and_rise * step/sweep_steps + start_pos
                
                # generate the gaussian
                Z = gaussian_2D(deg_x,deg_y,xy0[0],xy0[1],sigma_x,sigma_y,theta)
                Z_mask = np.zeros_like(Z)
                Z_mask[Z>0.33] = 1
                
                # loop over each flip
                for fr in np.arange(frames_per_vol):
                    
                    # get the frame number
                    f_num = tr_num * frames_per_vol + fr
                    
                    # set the frame to mean lum
                    bar_stimulus[:,:,f_num] = 128
                    
                    # get the bar pixels
                    xx,yy = np.nonzero(Z_mask)
                    
                    # set the amp modulation
                    bar_stimulus[xx,yy,f_num] = full_run[f_num]
                    
                # iterate TR
                tr_num += 1
                
        else:
            
            # step through each volume
            for step in np.arange(0,sweep_steps):
                                
                # step through each flip
                for f in np.arange(frames_per_vol):
                    
                    # get the frame number
                    f_num = tr_num * frames_per_vol + f
                    
                    # set the frame to mean luminance
                    bar_stimulus[:,:,f_num] = 128
                    
                # iterate
                tr_num += 1
       
    return bar_stimulus

def simulate_bar_stimulus(pixels_across, pixels_down, 
                          viewing_distance, screen_width, 
                          thetas, num_bar_steps, num_blank_steps, 
                          ecc, clip = 0.33):

    """
    A utility function for creating a sweeping bar stimulus in memory.
    
    This function creates a standard retinotopic mapping stimulus, the 
    sweeping bar. The user specifies some of the display and stimulus
    parameters. This is particularly useful for writing tests and simulating
    responses of visually driven voxels.
    
    pixels_across : int
        The number of pixels along the horizontal dimension of the display.
        
    pixels_down : int
        The number of pixels along the vertical dimension of the display.
    
    viewing_distance : float
        The distance between the participant and the display (cm).
    
    screen_width : float
        The width of the display (cm). This is used to compute the visual angle
        for determining the pixels per degree of visual angle.
    
    thetas : array-like
        An array containing the orientations of the bars that will sweep
        across the display.  For example `thetas = np.arange(0,360,360/8)`.
    
    num_steps : int
        The number of steps the bar makes on each sweep through the visual field.
    
    ecc : float
        The distance from fixation each bar sweep begins and ends (degees).
    
    blanks : bool
        A flag determining whether there are blank periods inserted at the beginning
        and the end of the stimulus run.
    
    clip : float
        The bar stimulus is created by cliping a very oblong two-dimensional
        Gaussian oriented orthogonally to the direction of the sweep.
        
    Returns
    -------
    bar : ndarray
        An array containing the bar stimulus. The array is three-dimensional, with the
        first two dimensions representing the size of the display and the third
        dimension representing time.
    
    
    """
    
    
    # visuotopic stuff
    ppd = np.pi*pixels_across/np.arctan(screen_width/viewing_distance/2.0)/360.0 # degrees of visual angle
    deg_x, deg_y = generate_coordinate_matrices(pixels_across, pixels_down, ppd, 1.0)
    
    # initialize the stimulus array
    total_trs = len(thetas[thetas==-1])*num_blank_steps + len(thetas[thetas!=-1])*num_bar_steps
    bar_stimulus = np.zeros((pixels_down, pixels_across, total_trs))
    
    # initialize a counter
    tr_num = 0
    
    # main loop
    for theta in thetas:
        
        if theta != -1:  # pragma: no cover
            
            # convert to radians
            theta_rad = theta * np.pi / 180
            
            # get the starting point and trajectory
            start_pos = np.array([-np.cos(theta_rad)*ecc, -np.sin(theta_rad)*ecc])
            end_pos = np.array([np.cos(theta_rad)*ecc, np.sin(theta_rad)*ecc])
            run_and_rise = end_pos - start_pos;
            
            if np.mod(theta,90) == 0:
                sigma_x = 1
                sigma_y = 100
            else:
                sigma_x = 100
                sigma_y = 1
            
            # step through each position along the trajectory
            for step in np.arange(0,num_bar_steps):
                
                # get the position of the bar at each step
                xy0 = run_and_rise * step/num_bar_steps + start_pos
                
                # generate the gaussian
                Z = gaussian_2D(deg_x,deg_y,xy0[0],xy0[1],sigma_x,sigma_y,theta)
                                
                # store and iterate
                bar_stimulus[:,:,tr_num] = Z
                tr_num += 1
                
        else:  # pragma: no cover
            for step in np.arange(0,num_blank_steps):
                tr_num += 1
                
    
    # digitize the bar stimulus
    bar = np.zeros_like(bar_stimulus)
    bar[bar_stimulus > clip] = 1
    bar = np.short(bar)
    
    return bar


# This should eventually be VisualStimulus, and there would be an abstract class layer
# above this called Stimulus that would be generic for n-dimentional feature spaces.
class VisualStimulus(StimulusModel):
    
    
    def __init__(self, stim_arr, viewing_distance, screen_width,
                 scale_factor, tr_length, dtype, interp='nearest'):
        
        """
        
        A child of the StimulusModel class for visual stimuli.
        
        
        Paramaters
        ----------
        
        stim_arr : ndarray
            An array containing the visual stimulus at the native resolution. The 
            visual stimulus is assumed to be three-dimensional (x,y,time).
        
        viewing_distance : float
            The distance between the participant and the display (cm).
            
        screen_width : float
            The width of the display (cm). This is used to compute the visual angle
            for determining the pixels per degree of visual angle.
        
        scale_factor : float
            The downsampling rate for ball=parking a solution. The `stim_arr` is
            downsampled so as to speed up the fitting procedure.  The final model
            estimates will be derived using the non-downsampled stimulus.
            
        """
        
        StimulusModel.__init__(self, stim_arr, dtype, tr_length)
        
        # absorb the vars
        self.viewing_distance = viewing_distance
        self.screen_width = screen_width
        self.scale_factor = scale_factor
        self.interp = interp
        
        # ascertain stimulus features
        self.pixels_across = self.stim_arr.shape[1]
        self.pixels_down = self.stim_arr.shape[0]
        self.run_length = self.stim_arr.shape[2]
        self.ppd = pixels_per_degree(self.pixels_across, self.screen_width, self.viewing_distance)
        
        # generate coordinate matrices
        deg_x, deg_y = generate_coordinate_matrices(self.pixels_across, self.pixels_down, self.ppd)
        
        # share coordinate matrices
        self.deg_x = utils.generate_shared_array(deg_x, ctypes.c_double)
        self.deg_y = utils.generate_shared_array(deg_y, ctypes.c_double)
        self.stim_arr = utils.generate_shared_array(stim_arr, dtype)
        
        if self.scale_factor == 1.0:
            
            self.stim_arr0 = self.stim_arr
            self.deg_x0 = self.deg_x
            self.deg_y0 = self.deg_y
            
        else:
            
            # create downsampled stimulus
            stim_arr0 = resample_stimulus(self.stim_arr, self.scale_factor)
            
            # generate the coordinate matrices
            deg_x0, deg_y0 = generate_coordinate_matrices(self.pixels_across, self.pixels_down, self.ppd, self.scale_factor)
            
            # share the arrays
            self.deg_x0 = utils.generate_shared_array(deg_x0, ctypes.c_double)
            self.deg_y0 = utils.generate_shared_array(deg_y0, ctypes.c_double)
            self.stim_arr0 = utils.generate_shared_array(stim_arr0, dtype)
        
        # add ppd for the down-sampled stimulus
        self.ppd0 = pixels_per_degree(self.pixels_across*self.scale_factor, self.screen_width, self.viewing_distance)
        
        



import ctypes

import matplotlib
from matplotlib.pyplot import specgram

import numpy as np
import numpy.testing as npt
from scipy.signal import chirp

import popeye.utilities as utils
import popeye.auditory_hrf as aud
from popeye.auditory_stimulus import AuditoryStimulus

def test_auditory_hrf_fit():
    
    # stimulus features
    duration = 30 # seconds
    Fs = int(44100/2) # Hz
    lo_freq = 200.0 # Hz
    hi_freq = 10000.0 # Hz
    tr_length = 1.0 # seconds
    clip_number = 0 # TRs
    dtype = ctypes.c_double
    
    # fit settings
    auto_fit = True
    verbose = 1
    debug = False
    Ns = 10
    
    # generate auditory stimulus
    time = np.linspace(0,duration,duration*Fs)
    ch = chirp(time, lo_freq, duration, hi_freq, method='logarithmic')
    signal = np.tile(np.concatenate((ch,ch[::-1])),5)
    blank = np.zeros((30*Fs))
    signal = np.concatenate((blank,signal,blank),-1)
    
    # instantiate an instance of the Stimulus class
    stimulus = AuditoryStimulus(signal, Fs, tr_length, dtype)  ### stimulus
    
    # initialize the gaussian model
    model = aud.AuditoryModel(stimulus, utils.spm_hrf)  ### model 
    model.hrf_delay = 0
    
    # invent pRF estimate
    center_freq_hz = 987
    sigma_hz = 123
    center_freq = np.log10(center_freq_hz)
    sigma = np.log10(sigma_hz)
    hrf_delay = 1.25
    beta = 2.4
    baseline = 0.59
    
    # generate data
    data = model.generate_prediction(center_freq, sigma, hrf_delay, beta, baseline)
    
    # search grids
    c_grid = utils.grid_slice(np.log10(300), np.log10(1000), Ns)
    s_grid = utils.grid_slice(np.log10(100), np.log10(500), Ns)
    h_grid = utils.grid_slice(-1,1,Ns)
    grids = (c_grid, s_grid, h_grid,)
    
    # search bounds
    c_bound = (np.log10(lo_freq), np.log10(hi_freq))
    s_bound = (np.log10(50), np.log10(hi_freq))
    h_bound = (-2,2)
    b_bound = (1e-8, None)
    m_bound = (None,None)
    bounds = (c_bound, s_bound, h_bound, b_bound, m_bound)
    
    # fit it
    fit = aud.AuditoryFit(model, data, grids, bounds, Ns=Ns)
    
    # grid fit
    npt.assert_almost_equal(fit.center_freq0, 3)
    npt.assert_almost_equal(fit.hrf0, 1.2222222222222223)
    # test the sigma parameter against best possibility
    grid_sigmas = np.arange(s_grid.start, s_grid.stop, s_grid.step)
    best_sigma = grid_sigmas[np.argmin(np.abs(grid_sigmas - sigma))]
    npt.assert_array_less(np.abs(fit.sigma0 - sigma), s_grid.step)
    # the baseline/beta should be 0/1 when regressed data vs. estimate
    (m,b) = np.polyfit(fit.scaled_ballpark_prediction, data, 1)
    npt.assert_almost_equal(m, 1.0)
    npt.assert_almost_equal(b, 0.0)
    
    # final fit
    npt.assert_almost_equal(fit.center_freq, center_freq)
    npt.assert_almost_equal(fit.sigma, sigma)
    npt.assert_almost_equal(fit.beta, beta)
    npt.assert_almost_equal(fit.baseline, baseline)
    npt.assert_almost_equal(fit.center_freq_hz, center_freq_hz)
    
    # test receptive field
    rf = np.exp(-((10**fit.model.stimulus.freqs-10**fit.center_freq)**2)/(2*(10**fit.sigma)**2))
    rf /= (10**fit.sigma*np.sqrt(2*np.pi))
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field.sum())) 
    
    # test model == fit RF
    rf = np.exp(-((fit.model.stimulus.freqs-fit.center_freq)**2)/(2*fit.sigma**2))
    rf /= (fit.sigma*np.sqrt(2*np.pi))
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field_log10.sum())) 
    
    



"""popeye version/release information"""

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"

_version_major = 1
_version_minor = 0
_version_micro = 0
# _version_extra = '.dev'
_version_extra = ''

# Construct full version string from these.
__version__ = "%s.%s.%s%s" % (_version_major,
                              _version_minor,
                              _version_micro,
                              _version_extra)

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: BSD License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

description = "popeye : voxel population receptive fields from fMRI data"

# Note: this long_description is actually a copy/paste from the top-level
# README.txt, so that it shows up nicely on PyPI.  So please remember to edit
# it only in one place and sync it correctly.
long_description = """
===============================================================
popeye : voxel population receptive fields from fMRI data
===============================================================

XXX Write a longer description here.


License information
===================

popeye is licensed under the terms of the new BSD license. See the file
"LICENSE" for information on the history of this software, terms & conditions
for usage, and a DISCLAIMER OF ALL WARRANTIES.

All trademarks referenced herein are property of their respective holders.

Copyright (c) 2013, Kevin DeSimone
All rights reserved.
"""

NAME = "popeye"
MAINTAINER = "Kevin DeSimone"
MAINTAINER_EMAIL = "kevindesimone@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "https://github.com/kdesimone/popeye/"
DOWNLOAD_URL = "http://github.com/kdesimone/popeye/downloads"
LICENSE = "Simplified BSD"
AUTHOR = "Kevin Desimone"
AUTHOR_EMAIL = "kevindesimone@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGES = ['popeye', 'popeye.tests']
PACKAGE_DATA = {"popeye": ["LICENSE"]}
REQUIRES = []



"""
Cross-validation analysis of population receptive field models
"""

from __future__ import division, print_function, absolute_import
from copy import deepcopy
import time
import numpy as np
import nibabel

def recast_xval_results(output, grid_parent, folds):
    
    # set dims for model+data
    dims = list(grid_parent.shape)
    dims.append(5)
    
    # initialize the cod array
    cod = np.zeros(dims)
    
    # extract the prf model estimates from the results queue output
    for o in output:
        xvoxel = o[0].voxel_index[0]
        yvoxel = o[0].voxel_index[1]
        zvoxel = o[0].voxel_index[2]
        cod[xvoxel,yvoxel,zvoxel,0] = np.mean((o[0].cod,o[1].cod))
        cod[xvoxel,yvoxel,zvoxel,1] = np.mean((o[0].rsquared,o[1].rsquared))
        cod[xvoxel,yvoxel,zvoxel,2] = np.mean((o[0].coefficient,o[1].coefficient))
        cod[xvoxel,yvoxel,zvoxel,3] = np.mean((o[0].mse,o[1].mse))
        cod[xvoxel,yvoxel,zvoxel,3] = np.mean((o[0].stderr,o[1].stderr))
        
    # get header information from the gridParent and update for the prf volume
    aff = grid_parent.get_affine()
    hdr = grid_parent.get_header()
    
    # recast as nifti
    nif = nibabel.Nifti1Image(cod,aff,header=hdr)
    nif.set_data_dtype('float32')
    
    return nif

def coeff_of_determination(data, model, axis=-1):
    
    r"""
    Calculate the coefficient of determination for a model prediction, relative
    to data.
    
    Parameters
    ----------
    data : ndarray
        The data
    model : ndarray
        The predictions of a model for this data. Same shape as the data.
    axis: int, optional
        The axis along which different samples are laid out (default: -1).
        
    Returns
    -------
    COD : ndarray
       The coefficient of determination.
    
    """
    
    residuals = data - model
    ss_err = np.sum(residuals ** 2, axis=axis)
    
    demeaned_data = data - np.mean(data, axis=axis)[..., np.newaxis]
    ss_tot = np.sum(demeaned_data **2, axis=axis)
    
    # Don't divide by 0:
    if np.all(ss_tot==0.0):
        return np.nan
        
    return 100 * (1 - (ss_err/ss_tot))

def kfold_xval(models, data, Fit, folds, fit_args, fit_kwargs):
    
    r"""
    Perform k-fold cross-validation to generate out-of-sample predictions for
    each measurement.
    
    Parameters
    ----------
    models : list of instances of Model
        A list containing the Model instances to be handed to Fit.  If the length of `models` is
        1, then it is assumed that `data` is composed of either a single run of data or of multiple
        runs with the same, repeated stimulus presented.  
    
    data : ndarray
        An m x n array representing a single voxel time-series, where m is the number of
        time-points and n is the number of runs
    
    Fit : Fit class object instance 
        The Fit class that will be instantiated with the left-in and left-out datasets.
    
    folds : int
        The number of divisions to apply to the data
        
    fit_args :
        Additional arguments to the model initialization
        
    fit_kwargs :
        Additional key-word arguments to the model initialization
        
    Notes
    -----
    This function assumes that a prediction API is implemented in the Fit
    class from which a prediction is conducted. That is, the Fit object that gets
    generated upon fitting the model needs to have a `prediction` method, which
    receives a functional time-series and a Model class instance as input 
    and produces a predicted signal as output.
    
    References
    ----------
    .. [1] Rokem, A., Chan, K.L. Yeatman, J.D., Pestilli, F., Mezer, A.,
       Wandell, B.A., 2014. Evaluating the accuracy of diffusion models at
       multiple b-values with cross-validation. ISMRM 2014.
    """
    
    
    # fold the data
    div_by_folds =  np.mod(data.shape[0], folds)
    
    # Make sure that an equal* number of samples get left out in each fold:
    if div_by_folds!= 0:
        msg = "The number of folds must divide the diffusion-weighted "
        msg += "data equally, but "
        msg = "np.mod(%s, %s) is %s"%(data.shape[0], folds, div_by_folds)
        raise ValueError(msg)
    
    # number of samples per fold
    n_in_fold = data.shape[-1]/folds
    
    # iniitialize a prediciton. this may not be necessary in popeye
    prediction = np.zeros(data.shape[-1])
    
    # We are going to leave out some randomly chosen samples in each iteration
    order = np.random.permutation(data.shape[0])
    order = np.reshape(order,(folds,int(len(order)/folds)))
    
    # initilize a list of predictions
    fits = []
    
    # Do the thing
    for k in range(folds):
        
        # Select the timepoints for this fold
        fold_mask = np.zeros(data.shape[0], dtype=bool)
        fold_mask[order[k]] = 1
        
        # Grab left-in data
        left_in_data = data[fold_mask,:]
        
        # Grab left-out data
        left_out_data = data[~fold_mask,:]
        
        # If there is only 1 model specified, compute the mean time-series for the voxel
        if len(models) == 1:
            
            left_in_data = np.mean(left_in_data,axis=0)
            left_out_data = np.mean(left_out_data,axis=0)
            model = models[0]
        
        # initialize the left-in fit object
        ensemble = []
        ensemble.append(model)
        ensemble.append(left_in_data)
        ensemble.extend(fit_args)
        ensemble.extend(fit_kwargs.values())
        left_in_fit = Fit(*ensemble)
                
        # initialize the left-out fit object
        ensemble = []
        ensemble.append(model)
        ensemble.append(left_out_data)
        ensemble.extend(fit_args)
        ensemble.extend(fit_kwargs.values())
        left_out_fit = Fit(*ensemble)
        
        # run the left-in Fit
        left_in_fit.estimate;
        
        # assign that estimate to the left out fit
        left_out_fit.estimate = left_in_fit.estimate
        
        # predict the left-out data
        left_out_fit.prediction;
        left_out_fit.dof = fold_mask.sum()-1
        
        
        # store the left-out data
        fits.append(left_out_fit)
    
    return fits


def parallel_xval(args):
    
    r"""
    This is a convenience function for parallelizing the fitting
    procedure.  Each call is handed a tuple or list containing
    all the necessary inputs for instantiaing a `GaussianFit`
    class object and estimating the model parameters.
    
    
    Paramaters
    ----------
    args : list/tuple
        A list or tuple containing all the necessary inputs for fitting
        the kfold_xval method.
    
    Returns
    -------
    
    fits : `PopulationFit` class objects that have been cross-validated.
        
    
    """ 
    
    # start timestamp
    tic = time.clock()
    
    # unpackage the arguments
    models = args[0]
    data = args[1]
    Fit = args[2]
    folds = args[3]
    fit_args = args[4]
    fit_kwargs = args[5]
    voxel_index = args[6]
    
    # cross-validate
    xval_fits =  kfold_xval(models, data, Fit, folds, fit_args, fit_kwargs)
    
    # compute the cod
    for fit in xval_fits:
        fit.cod = coeff_of_determination(fit.data, fit.prediction)
        fit.voxel_index = voxel_index
        fit.mse = np.std(fit.data)/np.sqrt(fit.dof)
    
    # end timestamp
    toc = time.clock()
    
    # print statement
    xvox = voxel_index[0]
    yvox = voxel_index[1]
    zvox = voxel_index[2]
    duration = toc-tic
    mean_cod = (xval_fits[0].cod+xval_fits[1].cod)/2/100
    mean_mse = (xval_fits[0].mse+xval_fits[1].mse)/2
    mean_bco = (xval_fits[0].coefficient+xval_fits[1].coefficient)/2
    mean_stderr = (xval_fits[0].stderr+xval_fits[1].stderr)/2
    print("Finished voxel %.03d,%.03d,%.03d in %.03d seconds  COD=%.02f BETACOEFF=%.02f  MSE=%.02f  STDERR=%.02f" % (xvox,yvox,zvox,duration,mean_cod,mean_bco,mean_mse,mean_stderr))
    return xval_fits
    
    


#!/usr/bin/python

""" Classes and functions for fitting SpatioTemporal population receptive field models """

from __future__ import division, print_function, absolute_import
import time
import gc
import warnings
warnings.simplefilter("ignore")

import numpy as np
from scipy.stats import linregress
from scipy.signal import fftconvolve
from scipy.integrate import trapz, simps
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_og_receptive_field, generate_mp_timeseries, generate_rf_timeseries

# Python 3 compatibility:
try:
    xrange
except NameError:  # pragma: no cover
    xrange = range


class SpatioTemporalModel(PopulationModel):
    
    """
    A Gaussian population receptive field model class
    
    """
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change):
        
        """
        A spatiotemporal population receptive field model.
        
        Paramaters
        ----------
        
        stimulus : `VisualStimulus` class object
            A class instantiation of the `VisualStimulus` class
            containing a representation of the visual stimulus.
            
            
        """
        
        PopulationModel.__init__(self, stimulus, hrf_model, normalizer)
        
        
    # for the final solution, we use spatiotemporal
    def generate_ballpark_prediction(self, x, y, sigma, n, weight):
        
        # mask for speed
        mask = self.distance_mask_coarse(x, y, sigma)
        
        # generate the RF
        spatial_rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x0, self.stimulus.deg_y0)
        spatial_rf /= ((2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x0[0,0:2])**2)
        
        # spatial_response
        spatial_ts = generate_rf_timeseries(self.stimulus.stim_arr0, spatial_rf, mask)
        
        # compression
        spatial_ts **= n
        
        # temporal response
        m_ts, p_ts = generate_mp_timeseries(spatial_ts, self.m_amp, self.p_amp, self.stimulus.flicker_vec)
        
        # mix them
        mp_ts = (1-weight) * m_ts + weight * p_ts
        
        # convolve with HRF
        model = fftconvolve(mp_ts, self.hrf())[0:len(mp_ts)]
        
        # units
        # model = (model - np.mean(model)) / np.mean(model)
        model = self.normalizer(model)
        
        # regress out mean and linear
        p = linregress(model, self.data)
        
        # scale
        model *= p[0]
        
        # offset
        model += p[1]
        
        return model
        
    # for the final solution, we use spatiotemporal
    def generate_prediction(self, x, y, sigma, n, weight, beta, baseline, unscaled=False):
        
        # mask for speed
        mask = self.distance_mask(x, y, sigma)
        
        # generate the RF
        spatial_rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        spatial_rf /= ((2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x[0,0:2])**2)
        
        # spatial response
        spatial_ts = generate_rf_timeseries(self.stimulus.stim_arr, spatial_rf, mask)
        
        # compression
        spatial_ts **= n
        
        # temporal response
        m_ts, p_ts = generate_mp_timeseries(spatial_ts, self.m_amp, self.p_amp, self.stimulus.flicker_vec)
        
        # mix them
        mp_ts = (1-weight) * m_ts + weight * p_ts
        
        # convolve with HRF
        model = fftconvolve(mp_ts, self.hrf())[0:len(mp_ts)]
        
        # convert units
        model = self.normalizer(model)
        
        if unscaled:
            return model
        else:
            # scale
            model *= beta
            
            # offset
            model += baseline
            
            return model

    @auto_attr
    def p(self):
        p = np.exp(-((self.t-self.center)**2)/(2*self.tau**2))
        p = p * 1/(np.sqrt(2*np.pi)*self.tau)
        return p
        
    @auto_attr
    def m(self):
        m = np.insert(np.diff(self.p),0,0)
        m = m/(simps(np.abs(m),self.t))
        return m

    def p_rf(self, tau):
        p = np.exp(-((self.t-self.center)**2)/(2*tau**2))
        p = p * 1/(np.sqrt(2*np.pi)*tau)
        return p

    def m_rf(self, tau):
        p = self.p_rf(tau)
        m = np.insert(np.diff(p),0,0)
        m = m/(simps(np.abs(m),self.t))
        return m

    @auto_attr
    def t(self):
        return np.linspace(0, self.stimulus.tr_length, self.stimulus.fps * self.stimulus.tr_length)

    @auto_attr
    def center(self):
        return self.stimulus.tr_length/2

    @auto_attr
    def flickers(self):
        return np.sin(2 * np.pi * np.single(self.stimulus.flicker_hz) * self.t[:,np.newaxis])

    @auto_attr
    def m_resp(self):
        m_resp = fftconvolve(self.flickers,self.m[:,np.newaxis])
        m_resp= utils.normalize(m_resp,-1,1)
        return m_resp

    def generate_m_resp(self, tau):
        m_rf = self.m_rf(tau)
        m_resp = fftconvolve(self.flickers,m_rf[:,np.newaxis])
        m_resp = utils.normalize(m_resp,-1,1)
        return m_resp

    @auto_attr
    def m_amp(self):
        m_amp = np.sum(np.abs(self.m_resp),0)
        m_amp /= m_amp.max()
        return m_amp

    @auto_attr
    def p_resp(self):
        p_resp = fftconvolve(self.flickers,self.p[:,np.newaxis])
        p_resp = utils.normalize(p_resp,-1,1)
        return p_resp

    def generate_p_resp(self, tau):
        p_rf = self.p_rf(tau)
        p_resp = fftconvolve(self.flickers,p_rf[:,np.newaxis])
        p_resp = utils.normalize(p_resp,-1,1)
        return p_resp

    @auto_attr
    def p_amp(self):
        p_amp = np.sum(np.abs(self.p_resp),0)
        p_amp /= np.max(p_amp)
        return p_amp


class SpatioTemporalFit(PopulationFit):
    
    """
    A spatiotemporal population receptive field fit class
    
    """
    
    def __init__(self, model, data, grids, bounds, 
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=0):
        
        PopulationFit.__init__(self, model, data, grids, bounds, 
                               voxel_index, Ns, auto_fit, verbose)
                               
    @auto_attr
    def overloaded_estimate(self):
        return [self.theta, self.rho, self.sigma_size, self.n, self.weight, self.beta]
    
    @auto_attr
    def x0(self):
        return self.ballpark[0]
        
    @auto_attr
    def y0(self):
        return self.ballpark[1]
        
    @auto_attr
    def sigma0(self):
        return self.ballpark[2]
        
    @auto_attr
    def n0(self):
        return self.ballpark[3]
        
    @auto_attr
    def weight0(self):
        return self.ballpark[4]
        
    @auto_attr
    def beta0(self):
        return self.ballpark[5]
        
    @auto_attr
    def baseline0(self):
        return self.ballpark[6]
        
    @auto_attr
    def x(self):
        return self.estimate[0]
        
    @auto_attr
    def y(self):
        return self.estimate[1]
        
    @auto_attr
    def sigma(self):
        return self.estimate[2]
        
    @auto_attr
    def n(self):
        return self.estimate[3]
        
    @auto_attr
    def weight(self):
        return self.estimate[4]
        
    @auto_attr
    def beta(self):
        return self.estimate[5]
    
    @auto_attr
    def baseline(self):
        return self.estimate[6]
        
    @auto_attr
    def rho(self):
        return np.sqrt(self.x**2+self.y**2)
        
    @auto_attr
    def theta(self):
        return np.mod(np.arctan2(self.y,self.x),2*np.pi)
        
    @auto_attr
    def sigma_size(self):
        return self.sigma / np.sqrt(self.n)
        
    @auto_attr
    def receptive_field(self):
        rf =  generate_og_receptive_field(self.x, self.y, self.sigma,
                                          self.model.stimulus.deg_x,
                                          self.model.stimulus.deg_y)
                                          
        rf /= ((2 * np.pi * self.sigma**2) * 1/np.diff(self.model.stimulus.deg_x0[0,0:2])**2)
        
        return rf




"""

Base-classes for poulation encoding models and fits.


"""
from popeye.onetime import auto_attr
import time, ctypes, itertools
import pickle
import sharedmem
from scipy.stats import linregress
import popeye.utilities as utils
import numpy as np
import numexpr as ne

try:  # pragma: no cover
    from types import SliceType
except ImportError:  # pragma: no cover
    SliceType = slice

def set_verbose(verbose):
    
    r"""A convenience function for setting the verbosity of a popeye model/fit.
    
    Paramaters
    ----------
    
    verbose : int
        0 = silent
        1 = print the final solution of an error-minimization
        2 = print each error-minimization step
        
    """
    
    # set verbose
    if verbose == 0: # pragma: no cover
        verbose = False
        very_verbose = False
    if verbose == 1: # pragma: no cover
        verbose = True
        very_verbose = False
    if verbose == 2: # pragma: no cover
        verbose = True
        very_verbose = True
    
    return verbose, very_verbose

class PopulationModel(object):
    
    r""" Base class for all pRF models."""
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change, cached_model_path=None, nuisance=None):
        
        r"""Base class for all pRF models.
        
        Paramaters
        ----------
        
        stimulus : `StimulusModel` class instance
            An instance of the `StimulusModel` class containing the `stim_arr` and 
            various stimulus parameters, and can represent various stimulus modalities, 
            including visual, auditory, etc.
        
        hrf_model : callable
            A function that generates an HRF model given an HRF delay.
            For more information, see `popeye.utilties.double_gamma_hrf_hrf
            and `popeye.utilities.spm_hrf`
            
        nuisance : ndarray
            A nuisance regressor for removing effects of non-interest.
            You can regress out any nuisance effects from you data prior to fitting
            the model of interest. The nuisance model is a statsmodels.OLS compatible
            design matrix, and the user is expected to have already added any constants.
        
        """
        
        # propogate
        self.stimulus = stimulus
        self.hrf_model = hrf_model
        self.normalizer = normalizer
        self.nuisance = nuisance
        self.cached_model_path = cached_model_path
        
        # set up cached model if specified
        if self.cached_model_path is not None: # pragma: no cover
            self.resurrect_cached_model
        
    def generate_ballpark_prediction(self): # pragma: no cover
        raise NotImplementedError("Each pRF model must implement its own ballpark prediction!") 
    
    def generate_prediction(self): # pragma: no cover
        raise NotImplementedError("Each pRF model must implement its own prediction!")
    
    def regress(self, X, y):
        slope, intercept = linregress(X, y)[0:2]
        if hasattr(self, 'bounded_amplitude') and self.bounded_amplitude:
            return np.abs(slope), intercept
        else:
            return slope, intercept
    
    def distance_mask_coarse(self, x, y, sigma):
        
        if hasattr(self, 'mask_size'): # pragma: no cover
            distance = (self.stimulus.deg_x0 - x)**2 + (self.stimulus.deg_y0 - y)**2
            mask = np.zeros_like(distance, dtype='uint8')
            mask[distance < self.mask_size*sigma**2] = 1
        else: # pragma: no cover
            mask = np.ones_like(self.stimulus.deg_x0, dtype='uint8')
            
        return mask
    
    def distance_mask(self, x, y, sigma):
        
        if hasattr(self, 'mask_size'): # pragma: no cover
            distance = (self.stimulus.deg_x - x)**2 + (self.stimulus.deg_y - y)**2
            mask = np.zeros_like(distance, dtype='uint8')
            mask[distance < self.mask_size*sigma**2] = 1
        else: # pragma: no cover
            mask = np.ones_like(self.stimulus.deg_x, dtype='uint8')
            
        return mask
    
    def hrf(self):
        if hasattr(self, 'hrf_delay'): # pragma: no cover
            return self.hrf_model(self.hrf_delay, self.stimulus.tr_length)
        else: # pragma: no cover
            raise NotImplementedError("You must set the HRF delay to generate the HRF")
    
    def cache_model(self, grids, ncpus=1, Ns=None, verbose=False):
        
        # get parameter space
        if isinstance(grids[0], SliceType):
            params = [list(np.arange(g.start,g.stop,g.step)) for g in grids]
        else:
            params = [list(np.linspace(g[0],g[1],Ns)) for g in grids]
        
        # make combos
        combos = np.array([c for c in itertools.product(*params)])
        # const = np.vstack((np.ones(combos.shape[0]),np.zeros(combos.shape[0]))).T
        # combos = np.concatenate((combos,const,1))
        
        # sort them by rand
        idx = np.argsort(np.random.rand(combos.shape[0]))
        combos = combos[idx]
        
        def mini_predictor(combo): # pragma: no cover
            print('%s' %(np.round(combo,2)))
            combo_long = list(combo)
            combo_long.extend((1,0))
            self.data = self.generate_prediction(*combo_long)
            return self.generate_ballpark_prediction(*combo), combo
        
        # compute predictions
        with sharedmem.Pool(np=ncpus) as pool:
            models = pool.map(mini_predictor, combos)
        
        # clean up
        models = [m for m in models if not np.isnan(np.sum(m[0]))]
        
        # turn into array
        return models
    
    @auto_attr
    def resurrect_cached_model(self):
        dat = pickle.load(open(self.cached_model_path, 'rb'))
        timeseries = utils.generate_shared_array(np.array([d[0] for d in dat]), np.double)
        parameters = utils.generate_shared_array(np.array([d[1] for d in dat]), np.double)
        return timeseries, parameters
    
    @auto_attr
    def cached_model_timeseries(self): # pragma: no cover
        return self.resurrect_cached_model[0]
    
    @auto_attr
    def cached_model_parameters(self): # pragma: no cover
        return self.resurrect_cached_model[1]
        
        
class PopulationFit(object):
    
    
    r""" Base class for all pRF model fits."""
    
    
    def __init__(self, model, data, grids, bounds, 
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=False):
        
        r"""A class containing tools for fitting pRF models.
        
        The `PopulationFit` class houses all the fitting tool that are associated with 
        estimatinga pRF model.  The `PopulationFit` takes a `PoulationModel` instance 
        `model` and a time-series `data`.  In addition, extent and sampling-rate of a 
        brute-force grid-search is set with `grids` and `Ns`.  Use `bounds` to set 
        limits on the search space for each parameter.  
        
        Paramaters
        ----------
        
        model : `PopulationModel` class instance
            An object representing pRF model.
        
        data : ndarray
            An array containing the measured BOLD signal of a single voxel.
        
        grids : tuple
            A tuple indicating the search space for the brute-force grid-search.
            The tuple contains pairs of upper and lower bounds for exploring a
            given dimension.  For example `grids=((-10,10),(0,5),)` will
            search the first dimension from -10 to 10 and the second from 0 to 5.
            These values cannot be `None`. 
            
            For more information, see `scipy.optimize.brute`.
        
        bounds : tuple
            A tuple containing the upper and lower bounds for each parameter
            in `parameters`.  If a parameter is not bounded, simply use
            `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would 
            bound the first parameter to be any positive number while the
            second parameter would be bounded between -10 and 10.
        
        voxel_index : tuple
            A tuple containing the index of the voxel being modeled. The 
            fitting procedure does not require a voxel index, but 
            collating the results across many voxels will does require voxel
            indices. With voxel indices, the brain volume can be reconstructed 
            using the newly computed model estimates.
        
        Ns : int
            Number of samples per stimulus dimension to sample during the ballpark search.
            
            For more information, see `scipy.optimize.brute`.
            
            Tnis option can be ignored if you want to specify your own sampling rate 
            for a given parameter. For more information, see `popeye.utilities.grid_slice`.
        
        auto_fit : bool
            A flag for automatically running the fitting procedures once the 
            `GaussianFit` object is instantiated.
        
        verbose : int
            0 = silent
            1 = print the final solution of an error-minimization
            2 = print each error-minimization step
        
        """
        
        # absorb vars
        self.grids = grids
        self.bounds = bounds
        self.Ns = Ns
        self.voxel_index = voxel_index
        self.auto_fit = auto_fit
        self.model = model
        self.data = data
        self.verbose, self.very_verbose = set_verbose(verbose)
        
        # regress out any nuisance
        if self.model.nuisance is not None: # pragma: no cover
            self.model.nuisance_model = sm.OLS(self.data,self.model.nuisance)
            self.model.results = self.model.nuisance_model.fit()
            self.original_data = self.data
            self.data = self.model.results.resid
        
        # push data down to model
        # there is probably a better way of exposing the data
        # to the PopulationModel. the idea is that the we want to
        # compute beta and baseline via linear regression rather
        # than estimate through optimization. Thus, model needs to see
        # the data. Since the data is in shared memory, no overhead incurred.
        self.model.data = data
        
        if self.bounds[-2][0] is not None and self.bounds[-2][0] > 0:
            self.model.bounded_amplitude = True # +/- amplitudes
        else:
            self.model.bounded_amplitude = False # + amplitudes
        
        # automatic fitting
        if self.auto_fit: # pragma: no cover
            
            # start
            self.start = time.time()
            
            # fit
            self.ballpark
            self.estimate
            self.overloaded_estimate
            
            # finish
            self.finish = time.time()
            
            # performance
            self.rss
            self.rsquared
            
            # flush if not testing
            if not hasattr(self.model, 'store_search_space'): # pragma: no cover
                self.gradient_descent = [None]*6
                self.brute_force = [None,]*4
            
            # print
            if self.verbose: # pragma: no cover
                print(self.msg)
    
    
    @auto_attr
    def best_cached_model_parameters(self): # pragma: no cover
        a = self.model.cached_model_timeseries
        b = self.data
        rss = ne.evaluate('sum((a-b)**2,axis=1)')
        idx = np.argmin(rss)
        return self.model.cached_model_parameters[idx]
    
    # the brute search
    @auto_attr
    def brute_force(self):
        return utils.brute_force_search(self.data,
                                        utils.error_function,
                                        self.model.generate_ballpark_prediction,
                                        self.grids,
                                        self.bounds,
                                        self.Ns,
                                        self.very_verbose)
    
    @auto_attr
    def ballpark(self):
        
        if self.model.cached_model_path is not None: # pragma: no cover
            return self.best_cached_model_parameters
        else:
            return np.append(self.brute_force[0],(self.slope,self.intercept))
    
    # the gradient search
    @auto_attr
    def gradient_descent(self):
        
        if self.very_verbose: # pragma: no cover
            print('The gridfit solution was %s, starting gradient descent ...' %(self.ballpark))
        
        return utils.gradient_descent_search(self.data,
                                             utils.error_function,
                                             self.model.generate_prediction,
                                             self.ballpark,
                                             self.bounds,
                                             self.very_verbose)
    @auto_attr
    def overloaded_estimate(self): # pragma: no cover
        
        """
        `overloaded_estimate` allows for representing the fitted
        parameter estimates in more useful units, such as 
        Cartesian to polar coordinates, or Hz to log(Hz).
        
        """
        
        return None
    
    @auto_attr
    def estimate(self):
        
        """
        `overloaded_estimate` allows for flexible representaiton of the
        final model estimate. For instance, you may fit the model parameters
        in cartesian space but would rather represent the fit in polar
        coordinates. 
        
        """
        
        return self.gradient_descent[0]
    
    @auto_attr
    def ballpark_prediction(self):
        return self.model.generate_prediction(*np.append(self.brute_force[0],(1,0)), unscaled=True)
    
    @auto_attr
    def scaled_ballpark_prediction(self):
        return self.ballpark_prediction * self.slope + self.intercept
    
    @auto_attr
    def slope(self):
        return self.model.regress(self.ballpark_prediction, self.data)[0]
    
    @auto_attr
    def intercept(self):
        return self.model.regress(self.ballpark_prediction, self.data)[1]
    
    @auto_attr
    def prediction(self):
        return self.model.generate_prediction(*self.estimate)
    
    @auto_attr
    def rsquared(self):
        return np.corrcoef(self.data, self.prediction)[1][0]**2
    
    @auto_attr
    def rss(self):
        return np.sum((self.data - self.prediction)**2)
    
    @auto_attr
    def msg(self):
        if self.auto_fit is True and self.overloaded_estimate is not None: # pragma: no cover
            txt = ("VOXEL=(%.03d,%.03d,%.03d)   TIME=%.03d   RSQ=%.02f  EST=%s"
                %(self.voxel_index[0],
                  self.voxel_index[1],
                  self.voxel_index[2],
                  self.finish-self.start,
                  self.rsquared,
                  np.round(self.overloaded_estimate,4)))
        elif self.auto_fit is True: # pragma: no cover
            txt = ("VOXEL=(%.03d,%.03d,%.03d)   TIME=%.03d   RSQ=%.02f  EST=%s"
                %(self.voxel_index[0],
                  self.voxel_index[1],
                  self.voxel_index[2],
                  self.finish-self.start,
                  self.rsquared,
                  np.round(self.estimate,4)))
        else: # pragma: no cover
            txt = ("VOXEL=(%.03d,%.03d,%.03d)   RSQ=%.02f  EST=%s" 
                %(self.voxel_index[0],
                  self.voxel_index[1],
                  self.voxel_index[2],
                  self.rsquared,
                  np.round(self.estimate,4)))
        return txt
            
class StimulusModel(object):

    def __init__(self, stim_arr, dtype=ctypes.c_int16, tr_length=1.0):
        
        r"""A base class for all encoding stimuli.
        
        This class houses the basic and common features of the encoding stimulus, which
        along with a `PopulationModel` constitutes what is commonly referred to as the 
        pRF model.
        
        Paramaters
        ----------
        
        stim_arr : ndarray
            An array containing the stimulus.  The dimensionality of the data is arbitrary
            but must be consistent with the pRF model, as specified in `PopulationModel` and 
            `PopluationFit` class instances.
        
        dtype : string
            Sets the data type the stimulus array is cast into.
        
        dtype : string
            Sets the data type the stimulus array is cast into.
        
        tr_length : float
            The repetition time (TR) in seconds.
            
        """
        
        self.dtype = dtype
        self.stim_arr = utils.generate_shared_array(stim_arr, self.dtype)
        self.tr_length = tr_length



#!/usr/bin/python

""" Classes and functions for fitting Gaussian population encoding models """

from __future__ import division, print_function, absolute_import
import time
import warnings
warnings.simplefilter("ignore")

import numpy as np
np.set_printoptions(suppress=True)
from scipy.stats import linregress
from scipy.signal import fftconvolve
from scipy.integrate import trapz

import nibabel
import numexpr as ne
from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_og_receptive_field, generate_rf_timeseries

class DifferenceOfGaussiansModel(PopulationModel):
    
    r"""
    A Difference of Gaussian population receptive field model class
    
    """
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change, cached_model_path=None, nuisance=None):
        
        r"""
        A Difference of Gaussian population receptive field model [1]_.
        
        Paramaters
        ----------
        
        stimulus : `VisualStimulus` class object
            A class instantiation of the `VisualStimulus` class
            containing a representation of the visual stimulus.
        
        hrf_model : callable
            A function that generates an HRF model given an HRF delay.
            For more information, see `popeye.utilties.double_gamma_hrf_hrf`
        
        References
        ----------
        
        .. [1] Zuiderbaan W, Harvey BM, Dumoulin SO (2012) Modeling 
        center-surroundconfigurations in population receptive fields 
        using fMRI. Journal of Vision 12(3):10,1-15.
        
        """
        PopulationModel.__init__(self, stimulus, hrf_model, normalizer)
        
        
    def generate_ballpark_prediction(self, x, y, sigma, sigma_ratio, volume_ratio):
        
        # extract the center response
        rf_center = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x0, self.stimulus.deg_y0)
        
        # extract surround response
        rf_surround = generate_og_receptive_field(x, y, sigma*sigma_ratio, 
                                                  self.stimulus.deg_x0, self.stimulus.deg_x0) * 1/sigma_ratio**2
        
        # difference
        rf = ne.evaluate('rf_center - sqrt(volume_ratio)*rf_surround')
        
        # extract the response
        mask = self.distance_mask(x, y, sigma*sigma_ratio)
        response = generate_rf_timeseries(self.stimulus.stim_arr0, rf, mask)
        
        # generate the hrf
        hrf = self.hrf_model(self.hrf_delay, self.stimulus.tr_length)
        
        # convolve it
        model = fftconvolve(response, hrf)[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        # regress out mean and linear
        beta, baseline = self.regress(model, self.data)
        
        # scale
        model *= beta
        
        # offset
        model += baseline
        
        return model
        
    def generate_prediction(self, x, y, sigma, sigma_ratio, volume_ratio, beta, baseline, unscaled=False):
        
        # extract the center response
        rf_center = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        
        # extract surround response
        rf_surround = generate_og_receptive_field(x, y, sigma*sigma_ratio, 
                                                  self.stimulus.deg_x, self.stimulus.deg_y) * 1/sigma_ratio**2
        
        # difference
        rf = ne.evaluate('rf_center - sqrt(volume_ratio)*rf_surround')
        
        # extract the response
        mask = self.distance_mask(x, y, sigma*sigma_ratio)
        response = generate_rf_timeseries(self.stimulus.stim_arr, rf, mask)
        
        # generate the hrf
        hrf = self.hrf_model(self.hrf_delay, self.stimulus.tr_length)
        
        # convolve it
        model = fftconvolve(response, hrf)[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        if unscaled:
            return model
        else:
            
            # scale it by beta
            model *= beta
            
            # offset
            model += baseline
            
            return model
            
    # DoG receptive field
    def receptive_field(self, x, y, sigma, sigma_ratio, volume_ratio):
            rf_center = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
            rf_surround = generate_og_receptive_field(x, y, sigma*sigma_ratio, self.stimulus.deg_x, self.stimulus.deg_y) * 1.0/sigma_ratio**2                                    
            rf = rf_center - np.sqrt(volume_ratio)*rf_surround
            return rf
    
class DifferenceOfGaussiansFit(PopulationFit):
    
    r"""
    A Difference of Gaussians population receptive field fit class
    
    """
    
    def __init__(self, model, data, grids, bounds,
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=0):
        
        r"""
        A class containing tools for fitting the Difference of Gaussians pRF model.
        
        The `DifferenceOfGaussiansFit` class houses all the fitting tool that 
        are associated with estimating a pRF model. The `GaussianFit` takes a 
        `DifferenceOfGaussiansModel` instance  `model` and a time-series `data`. 
        In addition, extent and sampling-rate of a  brute-force grid-search is set 
        with `grids` and `Ns`.  Use `bounds` to set limits on the search space for 
        each parameter.  
        
        Paramaters
        ----------
        
                
        model : `DifferenceOfGaussiansModel` class instance
            An object representing the CSS model. 
        
        data : ndarray
            An array containing the measured BOLD signal of a single voxel.
        
        grids : tuple
            A tuple indicating the search space for the brute-force grid-search.
            The tuple contains pairs of upper and lower bounds for exploring a
            given dimension.  For example `grids=((-10,10),(0,5),)` will
            search the first dimension from -10 to 10 and the second from 0 to 5.
            These values cannot be `None`. 
            
            For more information, see `scipy.optimize.brute`.
        
        bounds : tuple
            A tuple containing the upper and lower bounds for each parameter
            in `parameters`.  If a parameter is not bounded, simply use
            `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would 
            bound the first parameter to be any positive number while the
            second parameter would be bounded between -10 and 10.
        
        Ns : int
            Number of samples per stimulus dimension to sample during the ballpark search.
            
            For more information, see `scipy.optimize.brute`.
        
        voxel_index : tuple
            A tuple containing the index of the voxel being modeled. The 
            fitting procedure does not require a voxel index, but 
            collating the results across many voxels will does require voxel
            indices. With voxel indices, the brain volume can be reconstructed 
            using the newly computed model estimates.
        
        auto_fit : bool
            A flag for automatically running the fitting procedures once the 
            `GaussianFit` object is instantiated.
        
        verbose : int
            0 = silent
            1 = print the final solution of an error-minimization
            2 = print each error-minimization step
        
        """
        
        PopulationFit.__init__(self, model, data, grids, bounds, 
                               voxel_index, Ns, auto_fit, verbose)
    
    
    @auto_attr
    def overloaded_estimate(self):
       return [self.theta, self.rho, self.sigma, self.sigma_ratio, self.volume_ratio, self.beta, self.baseline] 
    
    @auto_attr
    def x0(self):
        return self.ballpark[0]
        
    @auto_attr
    def y0(self):
        return self.ballpark[1]
    
    @auto_attr
    def s0(self):
        return self.ballpark[2]
    
    @auto_attr
    def sr0(self):
        return self.ballpark[3]
    
    @auto_attr
    def vr0(self):
        return self.ballpark[4]
    
    @auto_attr
    def beta0(self):
        return self.ballpark[5]
    
    @auto_attr
    def baseline0(self):
        return self.ballpark[6]
    
    @auto_attr
    def x(self):
        return self.estimate[0]
        
    @auto_attr
    def y(self):
        return self.estimate[1]
        
    @auto_attr
    def sigma(self):
        return self.estimate[2]
        
    @auto_attr
    def sigma_ratio(self):
        return self.estimate[3]
    
    @auto_attr
    def volume_ratio(self):
        return self.estimate[4]
    
    @auto_attr
    def beta(self):
        return self.estimate[5]
    
    @auto_attr
    def baseline(self):
        return self.estimate[6]
    
    @auto_attr
    def rho(self):
        return np.sqrt(self.x**2+self.y**2)
    
    @auto_attr
    def theta(self):
        return np.mod(np.arctan2(self.y,self.x),2*np.pi)



import os
import sharedmem
import ctypes

import numpy as np
import numpy.testing as npt
import nose.tools as nt
from scipy.signal import fftconvolve

import popeye.utilities as utils
import popeye.og_hrf as og
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus
from popeye.spinach import generate_og_receptive_field

def test_og_fit():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    hrf_delay = 0.66
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, hrf_delay, beta, baseline)
    
    # set search grid
    x_grid = (-10,10)
    y_grid = (-10,10)
    s_grid = (0.25,5.25)
    h_grid = (-1.0,1.0)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    h_bound = (-1.5,1.5)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid, h_grid,)
    bounds = (x_bound, y_bound, s_bound, h_bound, b_bound, m_bound)
    
    # fit the response
    fit = og.GaussianFit(model, data, grids, bounds, Ns=5)
    
    # coarse fit
    npt.assert_almost_equal(fit.x0, -5.0)
    npt.assert_almost_equal(fit.y0, 5.0)
    npt.assert_almost_equal(fit.s0, 2.75)
    npt.assert_almost_equal(fit.hrf0, 0.5)
    # the baseline/beta should be 0/1 when regressed data vs. estimate
    (m,b) = np.polyfit(fit.scaled_ballpark_prediction, data, 1)
    npt.assert_almost_equal(m, 1.0)
    npt.assert_almost_equal(b, 0.0)

    # assert equivalence
    npt.assert_almost_equal(fit.x, x)
    npt.assert_almost_equal(fit.y, y)
    npt.assert_almost_equal(fit.hrf_delay, hrf_delay)
    npt.assert_almost_equal(fit.sigma, sigma)
    npt.assert_almost_equal(fit.beta, beta)
    
    # test receptive field
    rf = generate_og_receptive_field(x, y, sigma, fit.model.stimulus.deg_x, fit.model.stimulus.deg_y)
    rf /= (2 * np.pi * sigma**2) * 1/np.diff(model.stimulus.deg_x[0,0:2])**2
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field.sum())) 
    
    # test model == fit RF
    npt.assert_almost_equal(np.round(fit.model.generate_receptive_field(x,y,sigma).sum()), np.round(fit.receptive_field.sum()))

# def test_og_nuisance_fit():
#     
#     # stimulus features
#     viewing_distance = 38
#     screen_width = 25
#     thetas = np.arange(0,360,90)
#     thetas = np.insert(thetas,0,-1)
#     thetas = np.append(thetas,-1)
#     num_blank_steps = 30
#     num_bar_steps = 30
#     ecc = 12
#     tr_length = 1.0
#     frames_per_tr = 1.0
#     scale_factor = 1.0
#     pixels_across = 200
#     pixels_down = 200
#     dtype = ctypes.c_int16
#     
#     # create the sweeping bar stimulus in memory
#     bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
#                                 screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
#                                 
#     # create an instance of the Stimulus class
#     stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
#     
#     # initialize the gaussian model
#     model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
#     model.hrf_delay = 0
#     
#     # generate a random pRF estimate
#     x = -5.24
#     y = 2.58
#     sigma = 0.98
#     beta = 2.5
#     baseline = 0.0
#     
#     # create the "data"
#     data = model.generate_prediction(x, y, sigma, beta, baseline)
#     
#     # create nuisance signal
#     step = np.zeros(len(data))
#     step[30:-30] = 1
#     
#     # add to data
#     data += step
#     
#     # create design matrix
#     nuisance = sm.add_constant(step)
#     
#     # recreate model with nuisance
#     model = og.GaussianModel(stimulus, utils.double_gamma_hrf, nuisance)
#     model.hrf_delay = 0
#     
#     # set search grid
#     x_grid = (-7,7)
#     y_grid = (-7,7)
#     s_grid = (0.25,3.25)
#     
#     # set search bounds
#     x_bound = (-10.0,10.0)
#     y_bound = (-10.0,10.0)
#     s_bound = (0.001,10.0)
#     b_bound = (1e-8,None)
#     m_bound = (None, None)
#     
#     # loop over each voxel and set up a GaussianFit object
#     grids = (x_grid, y_grid, s_grid,)
#     bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
#     
#     # fit the response
#     fit = og.GaussianFit(model, data, grids, bounds, Ns=3)
#     
#     # assert equivalence
#     nt.assert_almost_equal(fit.x, x, 1)
#     nt.assert_almost_equal(fit.y, y, 1)
#     nt.assert_almost_equal(fit.sigma, sigma, 1)
#     nt.assert_almost_equal(fit.beta, beta, 1)
    



#!/usr/bin/python

""" Classes and functions for estimating Gaussian pRF model """

from __future__ import division
import warnings
warnings.simplefilter("ignore")

import numpy as np
from scipy.signal import fftconvolve
from scipy.stats import linregress
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_og_receptive_field, generate_rf_timeseries

class GaussianModel(PopulationModel):
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change, cached_model_path=None, nuisance=None):
        
        r"""A 2D Gaussian population receptive field model [1]_.
        
        Paramaters
        ----------
        
        stimulus : `VisualStimulus` class object
            A class instantiation of the `VisualStimulus` class
            containing a representation of the visual stimulus.
            
        hrf_model : callable
            A function that generates an HRF model given an HRF delay.
            For more information, see `popeye.utilties.double_gamma_hrf_hrf`
        
        
        References
        ----------
        
        .. [1] Dumoulin SO, Wandell BA. (2008) Population receptive field 
        estimates in human visual cortex. NeuroImage 39:647-660
        
        """
        
        PopulationModel.__init__(self, stimulus, hrf_model, normalizer)
        
    # main method for deriving model time-series
    def generate_ballpark_prediction(self, x, y, sigma):
        
        r"""
        Predict signal for the Gaussian Model using the downsampled stimulus.
        The rate of stimulus downsampling is defined in `model.stimulus.scale_factor`.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
        
        y: float 
            Vertical location of the Gaussian RF.
        
        sigma: float
            Dipsersion of the Gaussian RF.
        
        """
        
        # mask for speed
        mask = self.distance_mask_coarse(x, y, sigma)
        
        # generate the RF
        rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x0, self.stimulus.deg_y0)
        rf /= (2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x0[0,0:2])**2
                
        # extract the stimulus time-series
        response = generate_rf_timeseries(self.stimulus.stim_arr0, rf, mask)
        
        # convolve it with the stimulus
        model = fftconvolve(response, self.hrf())[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        # regress out mean and amplitude
        beta, baseline = self.regress(model, self.data)
        
        # scale
        model *= beta
        
        # offset
        model += baseline
        
        return model
        
    # main method for deriving model time-series
    def generate_prediction(self, x, y, sigma, beta, baseline, unscaled=False):
        
        r"""
        Predict signal for the Gaussian Model.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
        
        y: float 
            Vertical location of the Gaussian RF.
        
        sigma: float
            Dipsersion of the Gaussian RF.
        
        beta : float
            Amplitude scaling factor to account for units.
        
        baseline: float
            Amplitude intercept to account for baseline.
        
        """
        
        # mask for speed
        mask = self.distance_mask(x, y, sigma)
        
        # generate the RF
        rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        rf /= (2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x[0,0:2])**2
        
        # extract the stimulus time-series
        response = generate_rf_timeseries(self.stimulus.stim_arr, rf, mask)
        
        # convolve it with the stimulus
        model = fftconvolve(response, self.hrf())[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        if unscaled:
            return model
        else:
            
            # scale it by beta
            model *= beta
            
            # offset
            model += baseline
            
            return model
    
    def generate_receptive_field(self, x, y, sigma):
        
        r"""
        Generate a Gaussian receptive field in stimulus-referred coordinates.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
            
        y: float 
            Vertical location of the Gaussian RF.
            
        sigma: float
            Dipsersion of the Gaussian RF.
            
        """
        
        rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        rf /= (2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x[0,0:2])**2
        
        return rf
        
class GaussianFit(PopulationFit):
    
    r"""
    A 2D Gaussian population receptive field fit class.
    
    """
    
    def __init__(self, model, data, grids, bounds,
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=0):
        
        r"""
        A class containing tools for fitting the 2D Gaussian pRF model.
        
        The `GaussianFit` class houses all the fitting tool that are associated with 
        estimatinga pRF model.  The `GaussianFit` takes a `GaussianModel` instance 
        `model` and a time-series `data`.  In addition, extent and sampling-rate of a 
        brute-force grid-search is set with `grids` and `Ns`.  Use `bounds` to set 
        limits on the search space for each parameter.  
        
        Paramaters
        ----------
        
        
        model : `GaussianModel` class instance
            An object representing the 2D Gaussian model. 
            
        data : ndarray
            An array containing the measured BOLD signal of a single voxel.
            
        grids : tuple or Slice Object
            A tuple indicating the search space for the brute-force grid-search.
            The tuple contains pairs of upper and lower bounds for exploring a
            given dimension.  For example `grids=((-10,10),(0,5),)` will
            search the first dimension from -10 to 10 and the second from 0 to 5.
            The resolution of this search space is set with the `Ns` argument. 
            For more information, see `scipy.optimize.brute`.
            
            Alternatively you can pass `grids` a Slice Object. If you do use this
            option, you do not need to specificy `Ns` in `GaussianFit`. See 
            `popeye.utilities.grid_slice` for more details.
            
        bounds : tuple
            A tuple containing the upper and lower bounds for each parameter
            in `parameters`.  If a parameter is not bounded, simply use
            `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would 
            bound the first parameter to be any positive number while the
            second parameter would be bounded between -10 and 10.
            
        voxel_index : tuple
            A tuple containing the index of the voxel being modeled. The 
            fitting procedure does not require a voxel index, but 
            collating the results across many voxels will does require voxel
            indices. With voxel indices, the brain volume can be reconstructed 
            using the newly computed model estimates.
            
        Ns : int
            Number of samples per stimulus dimension to sample during the ballpark search.
            For more information, see `scipy.optimize.brute`.
            
            This can be `None` if `grids` is a tuple of Slice Objects.
            
        auto_fit : bool
            A flag for automatically running the fitting procedures once the 
            `GaussianFit` object is instantiated.
            
        verbose : int
            0 = silent
            1 = print the final solution of an error-minimization
            2 = print each error-minimization step
            
        """
        
        PopulationFit.__init__(self, model, data, grids, bounds, 
                           voxel_index, Ns, auto_fit, verbose)
                           
    @auto_attr
    def overloaded_estimate(self):
        return [self.theta, self.rho, self.sigma, self.beta, self.baseline]
    
    @auto_attr
    def x0(self):
        return self.ballpark[0]
        
    @auto_attr
    def y0(self):
        return self.ballpark[1]
        
    @auto_attr
    def s0(self):
        return self.ballpark[2]
    
    @auto_attr
    def beta0(self):
        return self.ballpark[3]
        
    @auto_attr
    def baseline0(self):
        return self.ballpark[4]
            
    @auto_attr
    def x(self):
        return self.estimate[0]
        
    @auto_attr
    def y(self):
        return self.estimate[1]
        
    @auto_attr
    def sigma(self):
        return self.estimate[2]
    
    @auto_attr
    def beta(self):
        return self.estimate[3]
    
    @auto_attr
    def baseline(self):
        return self.estimate[4]
        
    @auto_attr
    def rho(self):
        
        r""" Returns the eccentricity of the fitted pRF. """
        
        return np.sqrt(self.x**2+self.y**2)
    
    @auto_attr
    def theta(self):
        
        r""" Returns the polar angle of the fitted pRF. """
        
        return np.mod(np.arctan2(self.y,self.x),2*np.pi)
    
    @auto_attr
    def receptive_field(self):
        
        r""" Returns the fitted Gaussian pRF. """
        
        return self.model.generate_receptive_field(self.x, self.y, self.sigma)
                                           



from __future__ import division
import ctypes, sharedmem, sys

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import numpy as np
import nose.tools as nt
import numpy.testing as npt

import nibabel

import popeye.utilities as utils
import popeye.spinach as spin
import popeye.og as og
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus, generate_coordinate_matrices

def test_bootstrap():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.array([-1, 0, 90, 180, 270, -1])
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # rng
    np.random.seed(2764932)
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.5, 3.25, 5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # pack multiple "runs"
    data = np.vstack((data,data,data))
    
    # make it a singular "voxel"
    data = np.reshape(data, (1,data.shape[0],data.shape[1]))
    
    # set bootstraps and resamples
    bootstraps = 2
    resamples =  np.array((2,))
    
    # make fodder
    bundle = utils.bootstrap_bundle(bootstraps, resamples, og.GaussianFit, model, data, grids, bounds, np.tile((1,2,3),(bootstraps,1)))
    
    # test
    for b in bundle:
        fit = utils.parallel_bootstrap(b)
        npt.assert_almost_equal(fit.rss,0)
        npt.assert_equal(fit.n_resamples,resamples[0])
        npt.assert_equal(np.sum(fit.resamples), np.sum(np.arange(resamples[0])))

def test_xval():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.array([-1, 0, 90, 180, 270, -1])
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # rng
    np.random.seed(2764932)
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.5, 3.25, 5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # pack multiple "runs"
    data = np.vstack((data,data))
    
    # make it a singular "voxel"
    data = np.reshape(data, (1,data.shape[0],data.shape[1]))
    
    # set bootstraps and resamples
    bootstraps = 2
    kfolds = 2
    
    # make fodder
    bundle = utils.xval_bundle(bootstraps, kfolds, og.GaussianFit, model, data, grids, bounds, np.tile((1,2,3),(3,1)))
    
    # test
    for b in bundle:
        fit = utils.parallel_xval(b)
        npt.assert_almost_equal(fit.rss,0)
        npt.assert_equal(fit.cod, 100.0)
        npt.assert_(np.all(fit.tst_data == fit.trn_data))
        npt.assert_(np.all(fit.tst_idx != fit.trn_idx))

def test_distance_mask():
    
    x = 0
    y = 0
    sigma = 1
    amplitude=100
    dx,dy = np.meshgrid(np.linspace(-50,50,100),np.linspace(-50,50,100))
    mask = utils.distance_mask(x,y,sigma,dx,dy)
    npt.assert_equal(np.sqrt(np.sum(mask)),2)
    npt.assert_equal(np.max(mask),1)
    mask = utils.distance_mask(x,y,sigma,dx,dy,amplitude)
    npt.assert_equal(np.max(mask),amplitude)


def test_grid_slice():
    
    # test this case
    from_1 = 5
    to_1 = 15
    from_2 = 0
    to_2 = 2
    Ns=5
    
    # set a parameter to estimate
    params = (10,1)
    
    # see if we properly tile the parameter space for Ns=2
    grid_1 = utils.grid_slice(from_1, to_1, Ns)
    grid_2 = utils.grid_slice(from_2, to_2, Ns)
    grids = (grid_1, grid_2)
    
    # unbounded
    bounds = ()
    
    # create a simple function to generate a response from the parameter
    func = lambda freq,offset: np.sin( np.linspace(0,1,1000) * 2 * np.pi * freq) + offset
    
    # create a "response"
    response = func(*params)
    
    # get the ball-park estimate
    p0 = utils.brute_force_search(response, utils.error_function, func, grids, bounds)
    
    # make sure we fit right
    npt.assert_equal(params, p0[0])
    
    # make sure we sliced it right
    npt.assert_equal(p0[2][0].min(),from_1)
    npt.assert_equal(p0[2][0].max(),to_1)
    npt.assert_equal(p0[2][1].min(),from_2)
    npt.assert_equal(p0[2][1].max(),to_2)
    
def test_recast_estimation_results():

    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,45)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-5,4,5)
    y_grid = utils.grid_slice(-5,7,5)
    s_grid = utils.grid_slice(1/stimulus.ppd,5.25,5)
    b_grid = utils.grid_slice(0.1,4.0,5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (1/stimulus.ppd,12.0)
    b_bound = (1e-8,1e2)
    m_bound = (None,None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # create 3 voxels of data
    all_data = np.array([data,data,data])
    indices = [(0,0,0),(0,0,1),(0,0,2)]
    
    # bundle the voxels
    bundle = utils.multiprocess_bundle(og.GaussianFit, model, all_data, grids, bounds, indices)
    
    # run analysis
    with sharedmem.Pool(np=3) as pool:
        output = pool.map(utils.parallel_fit, bundle)
        
    # create grid parent
    arr = np.zeros((1,1,3))
    grid_parent = nibabel.Nifti1Image(arr,np.eye(4,4))
    
    # recast the estimation results
    nif = utils.recast_estimation_results(output, grid_parent)
    dat = nif.get_data()
    
    # assert equivalence
    npt.assert_almost_equal(np.mean(dat[...,0]), x)
    npt.assert_almost_equal(np.mean(dat[...,1]), y)
    npt.assert_almost_equal(np.mean(dat[...,2]), sigma)
    npt.assert_almost_equal(np.mean(dat[...,3]), beta)
    npt.assert_almost_equal(np.mean(dat[...,4]), baseline)
    
    # recast the estimation results - OVERLOADED
    nif = utils.recast_estimation_results(output, grid_parent, True)
    dat = nif.get_data()
    
    # assert equivalence
    npt.assert_almost_equal(np.mean(dat[...,0]), np.arctan2(y,x),2)
    npt.assert_almost_equal(np.mean(dat[...,1]), np.sqrt(x**2+y**2),2)
    npt.assert_almost_equal(np.mean(dat[...,2]), sigma)
    npt.assert_almost_equal(np.mean(dat[...,3]), beta)
    npt.assert_almost_equal(np.mean(dat[...,4]), baseline)


def test_make_nifti():

    # make up a volume
    arr = np.zeros((3,3,2))
    arr[...,0] = 1
    arr[...,1] = 100

    # make a nifti
    nif = utils.make_nifti(arr)

    npt.assert_equal(arr.shape,nif.shape)
    npt.assert_equal(np.mean(arr[...,0]),1)
    npt.assert_equal(np.mean(arr[...,1]),100)
    npt.assert_equal(nif.get_affine(),np.eye(4,4))

    # same except hand it a grid_parent
    grid_parent = nibabel.Nifti1Image(arr,np.eye(4,4))

    # make a nifti
    nif = utils.make_nifti(arr, grid_parent)

    npt.assert_equal(arr.shape,nif.shape)
    npt.assert_equal(np.mean(arr[...,0]),1)
    npt.assert_equal(np.mean(arr[...,1]),100)
    npt.assert_equal(nif.get_affine(),np.eye(4,4))

def test_normalize():

    # 1D
    arr = np.linspace(0,1,100)
    lo = 100.0
    hi = 200.0
    arr_new = utils.normalize(arr, lo, hi)
    npt.assert_equal(np.min(arr_new), lo)
    npt.assert_equal(np.max(arr_new), hi)

    # 2D
    arr = np.tile(np.linspace(0,1,100),(10,1))
    lo = np.repeat(lo,10)
    hi = np.repeat(hi,10)
    arr_new = utils.normalize(arr, lo, hi)
    npt.assert_equal(np.min(arr_new), lo)
    npt.assert_equal(np.max(arr_new), hi)

def test_error_function():

    # create a parameter to estimate
    params = (10.0,)

    # set bounds
    bounds = ((0.0,20.0),)

    # set the verbose level 0 is silent, 1 is final estimate, 2 is each iteration
    verbose = 0

    # create a simple function to transform the parameters
    func = lambda freq: np.sin( np.linspace(0,1,1000) * 2 * np.pi * freq)

    # create a "response"
    response = func(params[0])

    # assert 0 error
    npt.assert_equal(utils.error_function(params, bounds, response, func, verbose),0)

    # assert parameter outside of bounds return inf
    params = (30.0,)
    npt.assert_equal(utils.error_function(params, bounds, response, func, verbose), np.inf)

    # test nan returns inf
    response = func(params)
    response[0] = np.nan
    err = utils.error_function(params, bounds, response, func, verbose)
    npt.assert_equal(err,np.inf)

def test_gradient_descent_search():

    # create a parameter to estimate
    params = (10,10)

    # set grids + bounds
    grids = ((0,20),(5,15))
    bounds = ()

    # set the verbose level 0 is silent, 1 is final estimate, 2 is each iteration
    verbose = 0

    # set the number of search samples
    Ns = 3

    # create a simple function to transform the parameters
    func = lambda freq, offset: np.sin( np.linspace(0,1,1000) * 2 * np.pi * freq) + offset

    # create a "response"
    response = func(*params)

    # get the fine estimate
    phat = utils.gradient_descent_search(response, utils.error_function, func, (8,8), bounds, verbose)

    # assert that the estimate is equal to the parameter
    npt.assert_almost_equal(params, phat[0])

def test_brute_force_search_manual_grids():

    # create a parameter to estimate
    params = (10,10)

    # we need to define some search bounds
    grid_1 = utils.grid_slice(5,15,5)
    grid_2 = utils.grid_slice(5,15,5)
    grids = (grid_1,grid_2,)
    bounds = ()

    # set the verbose level 0 is silent, 1 is final estimate, 2 is each iteration
    verbose = 0

    # create a simple function to transform the parameters
    func = lambda freq, offset: np.sin( np.linspace(0,1,1000) * 2 * np.pi * freq) + offset

    # create a "response"
    response = func(*params)

    # get the ball-park estimate
    p0 = utils.brute_force_search(response, utils.error_function, func, grids, bounds)

    # assert that the estimate is equal to the parameter
    npt.assert_equal(params, p0[0])

def test_brute_force_search():

    # create a parameter to estimate
    params = (10,10)

    # we need to define some search bounds
    grids = ((0,20),(5,15))

    # we don't need to specify bounds for the error function
    bounds = ()

    # set the number of grid samples for the coarse search
    Ns = 3

    # set the verbose level 0 is silent, 1 is final estimate, 2 is each iteration
    verbose = 0

    # create a simple function to transform the parameters
    func = lambda freq, offset: np.sin( np.linspace(0,1,1000) * 2 * np.pi * freq) + offset

    # create a "response"
    response = func(*params)

    # get the ball-park estimate
    p0 = utils.brute_force_search(response, utils.error_function, func, grids, bounds, Ns=Ns)

    # assert that the estimate is equal to the parameter
    npt.assert_equal(params, p0[0])


def test_double_gamma_hrf():

    # set the TR length ... this affects the HRF sampling rate ...
    tr_length = 1.0

    hrf_0 = utils.double_gamma_hrf(-1, tr_length)
    hrf_1 = utils.double_gamma_hrf(0, tr_length)
    hrf_2 = utils.double_gamma_hrf(1, tr_length)
    npt.assert_almost_equal(hrf_0.sum(), hrf_1.sum(), 3)
    npt.assert_almost_equal(hrf_0.sum(), hrf_2.sum(), 3)
    npt.assert_almost_equal(hrf_1.sum(), hrf_2.sum(), 3)

    hrf_0 = utils.double_gamma_hrf(-1, tr_length, integrator=None)
    hrf_1 = utils.double_gamma_hrf(0, tr_length, integrator=None)
    hrf_2 = utils.double_gamma_hrf(1, tr_length, integrator=None)
    npt.assert_array_less(hrf_0.sum(),hrf_1.sum())
    npt.assert_array_less(hrf_1.sum(),hrf_2.sum())
    npt.assert_array_less(hrf_0.sum(),hrf_2.sum())


def test_spm_hrf():

    # set the TR length ... this affects the HRF sampling rate ...
    tr_length = 1.0

    # compute the difference in area under curve for hrf_delays of -1 and 0
    diff_1 = np.abs(np.sum(utils.spm_hrf(-1, tr_length))-np.sum(utils.spm_hrf(0, tr_length)))

    # compute the difference in area under curver for hrf_delays of 0 and 1
    diff_2 = np.abs(np.sum(utils.spm_hrf(1, tr_length))-np.sum(utils.spm_hrf(0, tr_length)))

    npt.assert_almost_equal(diff_1, diff_2, 2)

def test_glover_hrf():

    # set the TR length ... this affects the HRF sampling rate ...
    tr_length = 1.0

    # compute the difference in area under curve for hrf_delays of -1 and 0
    diff_1 = np.abs(np.sum(utils.glover_hrf(-1, tr_length))-np.sum(utils.glover_hrf(0, tr_length)))

    # compute the difference in area under curver for hrf_delays of 0 and 1
    diff_2 = np.abs(np.sum(utils.glover_hrf(1, tr_length))-np.sum(utils.glover_hrf(0, tr_length)))

    npt.assert_almost_equal(diff_1, diff_2, 2)


def test_randomize_voxels():

    # set the dummy dataset size
    x_dim, y_dim, z_dim = 10, 10, 10

    # create a dummy dataset with the
    dat = np.random.rand(x_dim, y_dim, z_dim)

    # mask and grab the indices
    xi,yi,zi = np.nonzero(dat>0.75)

    # create a random vector to resort the voxel indices
    rand_vec = np.random.rand(len(xi))
    rand_ind = np.argsort(rand_vec)

    # resort the indices
    rand_xi, rand_yi, rand_zi = xi[rand_ind], yi[rand_ind], zi[rand_ind]

    # assert that all members of the original and resorted indices are equal
    npt.assert_equal(set(xi),set(rand_xi))
    npt.assert_equal(set(yi),set(rand_yi))
    npt.assert_equal(set(zi),set(rand_zi))

def test_zscore():

    x = np.array([[1, 1, 3, 3],
                  [4, 4, 6, 6]])

    z = utils.zscore(x)
    npt.assert_equal(x.shape, z.shape)

    #Default axis is -1
    npt.assert_equal(utils.zscore(x), np.array([[-1., -1., 1., 1.],
                                                      [-1., -1., 1., 1.]]))

    #Test other axis:
    npt.assert_equal(utils.zscore(x, 0), np.array([[-1., -1., -1., -1.],
                                                        [1., 1., 1., 1.]]))


    # Test the 1D case:
    x = np.array([1, 1, 3, 3])
    npt.assert_equal(utils.zscore(x), [-1, -1, 1, 1])


def test_percent_change():
    x = np.array([[99, 100, 101], [4, 5, 6]])
    p = utils.percent_change(x)

    npt.assert_equal(x.shape, p.shape)
    npt.assert_almost_equal(p[0, 2], 1.0)

    ts = np.arange(4 * 5).reshape(4, 5)
    ax = 0
    npt.assert_almost_equal(utils.percent_change(ts, ax), np.array(
        [[-100., -88.23529412, -78.94736842, -71.42857143, -65.2173913],
        [-33.33333333, -29.41176471, -26.31578947, -23.80952381, -21.73913043],
        [33.33333333,   29.41176471,   26.31578947,   23.80952381, 21.73913043],
        [100., 88.23529412, 78.94736842, 71.42857143, 65.2173913]]))

    ax = 1
    npt.assert_almost_equal(utils.percent_change(ts, ax), np.array(
        [[-100., -50., 0., 50., 100.],
         [-28.57142857, -14.28571429, 0., 14.28571429, 28.57142857],
          [-16.66666667, -8.33333333, 0., 8.33333333, 16.66666667],
          [-11.76470588, -5.88235294, 0., 5.88235294, 11.76470588]]))

def test_parallel_fit_Ns():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    Ns = 3
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # make 3 voxels
    all_data = np.array([data,data,data])
    num_voxels = data.shape[0]
    indices = [(1,2,3)]*3
    
    # set search grid
    x_grid = (-10,10)
    y_grid = slice(-10,10)
    s_grid = (0.25,5.25)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,1e2)
    m_bound = (None, None)
    
    # make grids+bounds for all voxels in the sample
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # fitting params
    auto_fit = True
    verbose = 1
    
    # bundle the voxels
    bundle = utils.multiprocess_bundle(og.GaussianFit, model, all_data, grids, bounds, indices, Ns=3)
    
    # run analysis
    with sharedmem.Pool(np=3) as pool:
        output = pool.map(utils.parallel_fit, bundle)
        
    # assert equivalence
    for fit in output:
        npt.assert_almost_equal(fit.x, x, 2)
        npt.assert_almost_equal(fit.y, y, 2)
        npt.assert_almost_equal(fit.sigma, sigma, 2)
        npt.assert_almost_equal(fit.beta, beta, 2)
        npt.assert_almost_equal(fit.baseline, baseline, 2)
        
def test_parallel_fit():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,45)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = slice(-5,4,5)
    y_grid = slice(-5,7,5)
    s_grid = slice(1/stimulus.ppd,5.25,5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (1/stimulus.ppd,12.0)
    b_bound = (1e-8,1e2)
    m_bound = (None, None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # make 3 voxels
    all_data = np.array([data,data,data])
    num_voxels = data.shape[0]
    indices = [(1,2,3)]*3
    
    # bundle the voxels
    bundle = utils.multiprocess_bundle(og.GaussianFit, model, all_data, grids, bounds, indices)
    
    fit = utils.parallel_fit(bundle[0])
    
    # assert equivalence
    npt.assert_almost_equal(fit.x, x, 2)
    npt.assert_almost_equal(fit.y, y, 2)
    npt.assert_almost_equal(fit.sigma, sigma, 2)
    npt.assert_almost_equal(fit.beta, beta, 2)
    npt.assert_almost_equal(fit.baseline, baseline, 2)

def test_parallel_fit_manual_grids():

    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,45)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.10
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance,
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
    model.hrf_delay = 0
    
    # generate a random pRF estimate
    x = -5.24
    y = 2.58
    sigma = 1.24
    beta = 2.5
    baseline = -0.25
    
    # create the "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # set search grid
    x_grid = slice(-5,4,5)
    y_grid = slice(-5,7,5)
    s_grid = slice(1/stimulus.ppd,5.25,5)
    b_grid = slice(0.1,4.0,5)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (1/stimulus.ppd,12.0)
    b_bound = (1e-8,1e2)
    m_bound = (None, None)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # make 3 voxels
    all_data = np.array([data,data,data])
    num_voxels = data.shape[0]
    indices = [(1,2,3)]*3
    
    # bundle the voxels
    bundle = utils.multiprocess_bundle(og.GaussianFit, model, all_data, grids, bounds, indices)
    
    # run analysis
    with sharedmem.Pool(np=3) as pool:
        output = pool.map(utils.parallel_fit, bundle)
        
    # assert equivalence
    for fit in output:
        npt.assert_almost_equal(fit.x, x, 2)
        npt.assert_almost_equal(fit.y, y, 2)
        npt.assert_almost_equal(fit.sigma, sigma, 2)
        npt.assert_almost_equal(fit.beta, beta, 2)
        npt.assert_almost_equal(fit.baseline, baseline, 2)
        
def test_gaussian_2D():
    
    # set some dummy display parameters
    pixels_across = 101
    pixels_down = 101
    ppd = 1.0
    scale_factor = 1.0
    
    # generate coordinates
    deg_x, deg_y = generate_coordinate_matrices(pixels_across,pixels_down,ppd,scale_factor)
    
    # generate 2D case
    G = utils.gaussian_2D(deg_x,deg_y,0,0,10,10,0)
    
    # generate 1D case
    gx = np.exp(-((deg_x[0,:]-0)**2)/(2*10**2))
    gy = np.exp(-((deg_y[:,0]-0)**2)/(2*10**2))
    
    # assertions
    npt.assert_equal(np.round(G[:,50],5),np.round(gx,5))
    npt.assert_equal(np.round(G[50,:],5),np.round(gy,5))
    
def test_cartes_to_polar():
    cartes = np.array([5,0]).astype('double')
    polar = utils.cartes_to_polar(cartes)
    npt.assert_equal(polar[...,0], 0)
    npt.assert_equal(polar[...,1], 5)
    
    cartes = np.array([-5,0]).astype('double')
    polar = utils.cartes_to_polar(cartes)
    npt.assert_equal(polar[...,0], np.pi)
    npt.assert_equal(polar[...,1], 5)
    
    cartes = np.array([0,5]).astype('double')
    polar = utils.cartes_to_polar(cartes)
    npt.assert_equal(polar[...,0], np.pi/2)
    npt.assert_equal(polar[...,1], 5)
    
    cartes = np.array([0,-5]).astype('double')
    polar = utils.cartes_to_polar(cartes)
    npt.assert_equal(polar[...,0], np.pi*3/2)
    npt.assert_equal(polar[...,1], 5)

def test_binner():
    
    signal = np.ones(10)
    times = np.linspace(0,1,10)
    bins = np.arange(-0.5,1.5,0.5)
    binned_signal = utils.binner(signal, times, bins)
    
    npt.assert_equal(len(binned_signal), len(bins)-2)
    npt.assert_equal(binned_signal,[5,5])

def test_find_files():
    
    f = open('/tmp/test_abc.txt', 'w')
    f.close()
    
    path = utils.find_files('/tmp/','test*.txt')
    
    npt.assert_equal(path[0],'/tmp/test_abc.txt')
    
def test_peakdet():
    
    ts = np.zeros(100)
    
    peaks = np.arange(0,100,20)
    troughs = np.arange(10,100,20)
    
    ts[peaks] = 1
    ts[troughs] = -1
    
    a,b = utils.peakdet(ts,0.5)
    
    npt.assert_equal(a[:,0], peaks)
    npt.assert_equal(b[:,0], troughs)
    npt.assert_equal(a[:,1], 1)
    npt.assert_equal(b[:,1], -1)
    
def test_coeff_of_determination():
    
    # make up some data and a model
    data = np.arange(0,100)
    model = np.arange(0,100)
    
    # compute cod
    cod = utils.coeff_of_determination(data,model)
    
    # assert
    npt.assert_equal(cod, 100)
    
    # nan test
    cod = utils.coeff_of_determination(np.zeros_like(model),model)
    npt.assert_equal(cod,np.nan)
    
# def test_OLS():
#
#     o = utils.ols(np.arange(100),np.arange(100))
#
#     npt.assert_equal(o.R2,1.0)
#     npt.assert_almost_equal(np.sum(o.e),0.0)
#     npt.assert_almost_equal(np.sum(o.se),0.0)
#     npt.assert_true(o.F == np.inf)
#     npt.assert_almost_equal(np.sum(o.b),1.0)
#     npt.assert_true(o.df_e == len(np.arange(100-2)))
#     npt.assert_almost_equal(o.p[1],0)
#     npt.assert_true(o.ll() == (2987.7752827161585, -59.715505654323174, -59.663402250603411))
#     npt.assert_true(o.nobs == 100)
#     omni_1 = o.omni()[0]
#     omni_2 = o.omni()[1]
#     npt.assert_almost_equal(omni_1,18.093297390235648)
#     npt.assert_almost_equal(omni_2, 0.00011778511003501986)
#     npt.assert_true(o.JB() == (5.0825725194665461,0.07876502232916649,0.16483617111543283,1.9458968022816807))
#     npt.assert_true(o.dw() == 0.0051450432267976026)
#



# import os
# import multiprocessing
# from itertools import repeat
# import ctypes
# 
# import numpy as np
# import numpy.testing as npt
# import nose.tools as nt
# import scipy.signal as ss
# 
# 
# import popeye.utilities as utils
# from popeye import og
# import popeye.xvalidation as xval
# from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus
# from popeye.spinach import generate_og_timeseries
# 
# def test_coeff_of_determination():
#     
#     # make up some data and a model
#     data = np.arange(0,100)
#     model = np.arange(0,100)
#     
#     # compute cod
#     cod = xval.coeff_of_determination(data,model)
#     
#     # assert
#     npt.assert_equal(cod, 100)
#
# def test_kfold_xval_repeated_runs():
# 
#     # stimulus features
#     viewing_distance = 38
#     screen_width = 25
#     thetas = np.arange(0,360,90)
#     num_blank_steps = 0
#     num_bar_steps = 30
#     ecc = 10
#     tr_length = 1.0
#     frames_per_tr = 1.0
#     scale_factor = 0.10
#     pixels_down = 100
#     pixels_across = 100
#     dtype = ctypes.c_int16
#     Ns = 3
#     voxel_index = (1,2,3)
#     auto_fit = True
#     verbose = 1
#     
#     num_runs = 4
#     folds = 2
#     
#     # create the sweeping bar stimulus in memory
#     bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
#                                 screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
#     
#     # create an instance of the Stimulus class
#     stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
#     
#     # set up bounds for the grid search
#     grids = ((-10,10),(-10,10),(0.25,5.25))
#     bounds = ((-12,12),(-12,12),(1/stimulus.ppd,12),(1e-8,None),(None,None))
#     
#     # set the grid smaples
#     Ns = 5
#     
#     # initialize the gaussian model
#     model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
#     model.hrf_delay = 0
#     
#     # generate a random pRF estimate
#     x = -5.24
#     y = 2.58
#     sigma = 1.24
#     beta = 2.5
#     baseline = -0.25
#     
#     # create the args context for calling the Fit class
#     fit_args = [grids, bounds, [0,0,0], Ns]
#     fit_kwargs = {'auto_fit': False, 'verbose' : 0}
#     
#     # create a series of "runs"
#     data = np.zeros((num_runs,stimulus.stim_arr.shape[-1]))
#     
#     for r in range(num_runs):
#         
#         # fill out the data list
#         data[r,:] = model.generate_prediction(x, y, sigma, beta, baseline)
#     
#     # get predictions out for each of the folds ...
#     models = (model,)
#     predictions = xval.kfold_xval(models, data, og.GaussianFit, folds, fit_args, fit_kwargs)
#     
#     # assert the coeff of determination is 100 for each prediction
#     for p in predictions:
#         cod = xval.coeff_of_determination(p.data,p.prediction)
#         npt.assert_almost_equal(cod,100, 4)

# def test_kfold_xval_unique_runs():
#     
#     # stimulus features
#     pixels_across = 800
#     pixels_down = 600
#     viewing_distance = 38
#     screen_width = 25
#     thetas = np.arange(0,360,45)
#     num_steps = 20
#     ecc = 10
#     tr_length = 1.0
#     frames_per_tr = 1.0
#     scale_factor = 0.05
#     dtype = ctypes.c_short
#     num_runs = 4
#     folds = 2
#     
#     # create the sweeping bar stimulus in memory
#     bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, thetas, num_steps, ecc)
#     
#     # create an instance of the Stimulus class
#     stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, dtype)
#     
#     # set up bounds for the grid search
#     grids = ((-10,10),(-10,10),(0.25,5.25),(0.1,1e2),(-5,5))
#     bounds = ((-12,12),(-12,12),(1/stimulus.ppd,12),(0.1,1e3),(-5,5))
#     
#     # initialize the gaussian model
#     model = og.GaussianModel(stimulus)
#     
#     # generate a random pRF estimate
#     x = -5.24
#     y = 2.58
#     sigma = 1.24
#     beta = 2.5
#     hrf_delay = -0.25
#     
#     # create the args context for calling the Fit class
#     fit_args = [grids, bounds, tr_length, [0,0,0],]
#     fit_kwargs = {'auto_fit': False, 'verbose' : False}
#     
#     # create a series of "runs"
#     data = np.zeros((stimulus.stim_arr.shape[-1],num_runs))
#     
#     for r in range(num_runs):
#         
#         # fill out the data list
#         data[:,r] = og.compute_model_ts(x, y, sigma, beta, hrf_delay,
#                                         stimulus.deg_x, stimulus.deg_y, 
#                                         stimulus.stim_arr, tr_length)
#     
#     
#     # get predictions out for each of the folds ...
#     models = np.tile(model,num_runs)
#     left_out_data, predictions = xval.kfold_xval(models, data, og.GaussianFit, folds, fit_args, fit_kwargs)
#     
#     # assert the coeff of determination is 100 for each prediction
#     for k in range(folds):
#         cod = xval.coeff_of_determination(left_out_data[k], predictions[k])
#         npt.assert_almost_equal(cod,100, 4)


import numpy as np
import numpy.testing as npt
import nose.tools as nt

import popeye.utilities as utils
import popeye.spinach as spin
from popeye.visual_stimulus import generate_coordinate_matrices
from popeye.spinach import generate_og_receptive_field, generate_og_timeseries

def test_generate_og_receptive_field():
    xpixels = 500 # simulated screen width
    ypixels = 500 # simulated screen height
    ppd = 1 # simulated visual angle
    scale_factor = 1.0 # simulated stimulus resampling rate
    distance = 5 # standard deviations to compute gauss out to
    xcenter = 0 # x coordinate of the pRF center
    ycenter = 0 # y coordinate of the pRF center
    sigma = 1 # width of the pRF
    
    test_value = 6 # this is the sum of a gaussian given 1 ppd 
                   # and a 1 sigma prf centered on (0,0)
                   
    # generate the visuotopic coordinates
    dx,dy = generate_coordinate_matrices(xpixels,
                                         ypixels,
                                         ppd,
                                         scale_factor)
    
    # generate a pRF at (0,0) and 1 sigma wide
    rf = generate_og_receptive_field(xcenter, ycenter, sigma, dx, dy)
    
    # divide by integral
    rf /= 2 * np.pi * sigma ** 2
    
    # compare the volume of the pRF to a known value
    nt.assert_almost_equal(np.sum(rf),1)


def test_generate_og_timeseries():
    xpixels = 100 # simulated screen width
    ypixels = 100 # simulated screen height
    ppd = 1 # simulated visual angle
    scaleFactor = 1.0 # simulated stimulus resampling rate
    
    xcenter = 0 # x coordinate of the pRF center
    ycenter = 0 # y coordinate of the pRF center
    sigma = 10 # width of the pRF
    
    # generate the visuotopic coordinates
    dx,dy = generate_coordinate_matrices(xpixels,
                                         ypixels,
                                         ppd,
                                         scaleFactor)
    
    timeseries_length = 15 # number of frames to simulate our stimulus array
    
    # initialize the stimulus array
    stim_arr = np.zeros((xpixels, ypixels, timeseries_length)).astype('uint8')
    
    # make a circular mask appear for the first 5 frames
    xi,yi = np.nonzero(np.sqrt((dx-xcenter)**2 + (dy-ycenter)**2)<sigma)
    stim_arr[xi,yi,0:5] = 1
    
    # make an annulus appear for the next 5 frames
    xi,yi = np.nonzero(np.sqrt((dx-xcenter)**2 + (dy-ycenter)**2)>sigma)
    stim_arr[xi,yi,5:10] = 1
    
    # make a circular mask appear for the next 5 frames
    xi,yi = np.nonzero(np.sqrt((dx-xcenter)**2 + (dy-ycenter)**2)<sigma)
    stim_arr[xi,yi,10::] = 1
    
    
    # make the response prediction
    response = generate_og_timeseries(dx,
                                      dy,
                                      stim_arr,
                                      xcenter,
                                      ycenter,
                                      sigma)
                                
    # make sure the RSS is 0
    step = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0])
    rval = np.corrcoef(response, step)[0, 1]
    
    # The voxel responds when the stimulus covers its PRF, so it perfectly
    # correlates with a step function: 
    nt.assert_equal(round(rval, 3), 1)

# def test_binner():
#     
#     signal = np.ones(10)
#     times = np.linspace(0,1,10)
#     bins = np.arange(-0.5,1.5,0.5)
#     
#     binned_signal = spin.binner(signal, times, bins)
#     
#     nt.assert_true(len(binned_signal), len(bins)-2)
#     nt.assert_true(np.all(binned_signal==[5,5]))
#     



# from __future__ import division
# 
# from itertools import repeat
# import multiprocessing
# 
# import numpy as np
# import numpy.testing as npt
# import nose.tools as nt
# 
# from scipy.signal import chirp
# 
# import popeye.utilities as utils
# import popeye.strf as strf
# from popeye.auditory_stimulus import AuditoryStimulus
# 
# # def test_brute_force_search():
# #     """
# #     Test voxel-wise strf estimation function in popeye.estimation 
# #     using the stimulus and BOLD time-series data that ship with the 
# #     popeye installation.
# #     """
# #     
# #     # stimulus features
# #     lo_freq = 100 # Hz
# #     hi_freq = 1000 # Hz
# #     fs = hi_freq*2 # Hz
# #     duration = 100 # seconds
# #     tr_length = 1.0 # seconds
# #     time_window = 0.5 # seconds
# #     freq_window = 256 # this is 2x the number of freq bins we'll end up with in the spectrogram
# #     
# #     num_timepoints = np.floor(duration / tr_length)
# #     
# #     # sample the time from 0-duration by the fs
# #     time = np.linspace(0,duration,duration*fs)
# #     
# #     # create the sweeping bar stimulus in memory
# #     signal = chirp(time, lo_freq, duration, hi_freq, method='linear')
# #     %
# #     # instantiate an instance of the Stimulus class
# #     stimulus = AuditoryStimulus(signal, tr_length, freq_window, time_window)
# #     
# #     # set up bounds for the grid search
# #     bounds = ((lo_freq, hi_freq),(lo_freq, hi_freq),(-5,5))
# #     
# #     # initialize the strf model
# #     strf_model = strf.SpectrotemporalModel(stimulus)
# #     
# #     # makeup a STRF estimate
# #     freq_center = 550 # center frequency
# #     freq_sigma = 100 # frequency dispersion
# #     hrf_delay = 0.0 # seconds
# #     
# #     # generate the modeled BOLD response
# #     response = strf.compute_model_ts(freq_center, freq_sigma, hrf_delay,
# #                                      stimulus.time_coord, stimulus.freq_coord, stimulus.spectrogram,
# #                                      tr_length, num_timepoints)
# #     
# #     # compute the initial guess with the adaptive brute-force grid-search
# #     f0, fs0, hrf0 = strf.brute_force_search(bounds, response,
# #                                             stimulus.time_coord,
# #                                             stimulus.freq_coord,
# #                                             stimulus.spectrogram,
# #                                             tr_length, num_timepoints)
# #                                                              
# #         
# #     # assert
# #     npt.assert_equal([freq_center, freq_sigma, hrf_delay], [f0,fs0,hrf0])
# 
# def test_strf_fit():
#     
#     # stimulus features
#     lo_freq = 200 # Hz
#     hi_freq = 12000 # Hz
#     fs = hi_freq*2 # Hz
#     duration = 100 # seconds
#     tr_length = 1.0 # seconds
#     tr_sampling_rate = 1 # number of time-samples per TR to plot the STRF in
#     time_window = 0.5 # seconds
#     freq_window = 256 # this is 2x the number of freq bins we'll end up with in the spectrogram
#     scale_factor = 1.0 # how much to downsample the spectrotemporal space
#     num_timepoints = np.floor(duration / tr_length)
#     degrees = 0.0
#     
#     # sample the time from 0-duration by the fs
#     time = np.linspace(0,duration,duration*fs)
#     
#     # create a chirp stimulus
#     signal = chirp(time, lo_freq, duration, hi_freq)
#     
#     # instantiate an instance of the Stimulus class
#     stimulus = AuditoryStimulus(signal, tr_length, freq_window, time_window, sampling_rate=fs, 
#                                 tr_sampling_rate=tr_sampling_rate, scale_factor=scale_factor)
#     
#     
#     # set some parameters for the mock STRF
#     freq_center = 5678 # center frequency
#     freq_sigma = 234 # frequency dispersion
#     hrf_delay = 0.987 # seconds
#     
#     # initialize the strf model
#     model = strf.SpectrotemporalModel(stimulus)
#     
#     # generate the modeled BOLD response
#     data = strf.compute_model_ts(freq_center, freq_sigma, hrf_delay,
#                                  model.stimulus.time_coord, 
#                                  model.stimulus.freq_coord,
#                                  model.stimulus.spectrogram.astype('double'),
#                                  tr_length, num_timepoints, norm_func=utils.zscore)
#     
#     # set some searh parameters
#     search_bounds = ((lo_freq, hi_freq),(lo_freq, hi_freq/2),(-5,5),)
#     fit_bounds = ((lo_freq, hi_freq),(lo_freq, hi_freq/2),(-5,5),)
#     
#     # fit the response
#     fit = strf.SpectrotemporalFit(data, model, search_bounds, fit_bounds, tr_length)
#     
#     # assert
#     npt.assert_almost_equal(fit.estimate,[freq_center,freq_sigma,hrf_delay])
# 
# def test_parallel_fit():
#     
#     lo_freq = 200 # Hz
#     hi_freq = 12000 # Hz
#     fs = 44100.0 # Hz
#     duration = 100 # seconds
#     tr_length = 1.0 # seconds
#     tr_sampling_rate = 10 # number of time-samples per TR to plot the STRF in
#     time_window = 0.5 # seconds
#     freq_window = 256 # this is 2x the number of freq bins we'll end up with in the spectrogram
#     scale_factor = 1.0 # how much to downsample the spectrotemporal space
#     num_timepoints = np.floor(duration / tr_length)
#     degrees = 0.0
#     num_voxels = multiprocessing.cpu_count()
#     
#     
#     # sample the time from 0-duration by the fs
#     time = np.linspace(0,duration,duration*fs)
#     
#     # create a chirp stimulus
#     signal = chirp(time, lo_freq, duration, hi_freq)
#     
#     # instantiate an instance of the Stimulus class
#     stimulus = AuditoryStimulus(signal, tr_length, freq_window, time_window, sampling_rate=fs, 
#                                 tr_sampling_rate=tr_sampling_rate, scale_factor=scale_factor)
#     
#     
#     # invent a pRF estimate
#     freq_center = 5678 # center frequency
#     freq_sigma = 234 # frequency dispersion
#     hrf_delay = 0.987 # seconds
#     
#     # initialize the strf model
#     model = strf.SpectrotemporalModel(stimulus)
#     
#     # set some bounds
#     search_bounds = [((lo_freq, hi_freq),(lo_freq, hi_freq/2),(-5,5),)]*num_voxels
#     fit_bounds = [((lo_freq, hi_freq),(lo_freq, hi_freq/2),(-5,5),)]*num_voxels
#     
#     # generate the modeled BOLD response
#     response = strf.compute_model_ts(freq_center, freq_sigma, hrf_delay,
#                                      model.stimulus.time_coord, 
#                                      model.stimulus.freq_coord,
#                                      model.stimulus.spectrogram.astype('double'),
#                                      tr_length, num_timepoints, norm_func=utils.zscore)
#     
#     
#     # package the data structure
#     dat = zip(repeat(response,num_voxels),
#               repeat(model,num_voxels),
#               search_bounds,
#               fit_bounds,
#               repeat(tr_length,num_voxels),
#               repeat(None,num_voxels),
#               repeat(True,num_voxels),
#               repeat(False,num_voxels))
#               
#     # run analysis
#     num_cpus = multiprocessing.cpu_count()-1
#     pool = multiprocessing.Pool(num_cpus)
#     output = pool.map(strf.parallel_fit,dat)
#     pool.close()
#     pool.join()
#     
#     # assert equivalence
#     for fit in output:
#         npt.assert_almost_equal(fit.estimate,[freq_center,freq_sigma,hrf_delay])
#     
